import torch
from .BaseContainer import BaseContainer
from .. import util

class FlatContainerGroup(BaseContainer):
  def __init__(
      self,
      code_size_list,
      dtype_list,
      device_list,
      contiguous_size_list=None,
      initial_size=None,
      expand_step_size=1024,
      expand_mode="double",
      use_inverse_id_mapping=False,
      verbose=0
    ):
    assert type(code_size_list) in [list, tuple]
    n_storage = len(code_size_list)
    assert type(device_list) in [list, tuple]
    assert type(dtype_list) in [list, tuple]
    assert len(device_list) == len(dtype_list) == n_storage
    dtype_list = [util.str2dtype(i) for i in dtype_list]
    if contiguous_size_list is not None:
      assert type(contiguous_size_list) in [list, tuple]
      assert len(contiguous_size_list) == n_storage
      for i in range(n_storage):
        assert code_size_list[i] % contiguous_size_list[i] == 0
    else:
      contiguous_size_list = [1] * n_storage

    main_device = "cuda:0" if "cuda" in [torch.device(i).type for i in device_list] else "cpu"
    super(FlatContainerGroup, self).__init__(
      device=main_device,
      initial_size=initial_size,
      expand_step_size=expand_step_size,
      expand_mode=expand_mode,
      use_inverse_id_mapping=use_inverse_id_mapping
    )

    self.code_size_list = code_size_list
    self.device_list = device_list
    self.dtype_list = dtype_list
    self.contiguous_size_list = contiguous_size_list
    self.verbose = verbose
    self._n_items = 0
    n_storage = len(self.code_size_list)
    for i in range(n_storage):
      storage = torch.zeros(
        code_size_list[i] // contiguous_size_list[i],
        self.initial_size,
        contiguous_size_list[i],
        dtype = dtype_list[i],
        device = device_list[i]
      )
      self.register_buffer(f"_storage_{i}", storage)

  @property
  def n_storage(self):
    return len(self.code_size_list)

  @property
  def n_items(self):
    return self._n_items

  def __getitem__(self, index):
    module = FlatContainer(
      code_size = self.code_size_list[index],
      expand_step_size = self.expand_step_size,
      initial_size = self.initial_size,
      contiguous_size = self.contiguous_size_list[index],
      use_inverse_id_mapping = self.use_inverse_id_mapping,
      expand_mode = self.expand_mode,
      verbose = self.verbose,
      dtype = self.dtype_list[index],
      device = self.device_list[index]
    )
    module._n_items = self._n_items
    del module._storage
    del module._address2id
    del module._id2address
    
    storage = getattr(self, f"_storage_{index}")
    module.register_buffer("_storage", storage)
    module.register_buffer("_id2address", self._id2address)
    module.register_buffer("_address2id", self._address2id)

    def read_only_alert(*args, **kwargs):
      raise NotImplementedError("Module cannot be modified")
    module.add = read_only_alert
    module.remove = read_only_alert
    module.expand = read_only_alert

    return module
    
  def get_data_by_address(self, address):
    assert util.check_dtype(address, torch.int64)
    address = address.to(self.device).clone()
    
    n_address = address.shape[0]
    mask = (0 <= address) & (address < self.capacity)
    address[~mask] = 0
    result = []
    for i in range(self.n_storage):
      device = self.device_list[i]
      dtype = self.dtype_list[i]
      code_size = self.code_size_list[i]
      storage = getattr(self, f"_storage_{i}")
      data = storage.index_select(
        dim=1,
        index=address.to(device)
      )
      data.index_fill_(
        dim=1,
        index=torch.nonzero(~mask).squeeze(1).to(device),
        value=0
      )
      data = data.transpose(1, 2).reshape(code_size, -1)
      result.append(data)
    return result

  def set_data_by_address(self, address, data):
    assert util.check_dtype(address, torch.int64)
    assert type(data) in [list, tuple]
    assert len(data) == self.n_storage
    address = address.to(self.device)
    mask = (0 <= address) & (address < self.capacity)

    for i in range(self.n_storage):
      device = self.device_list[i]
      dtype = self.dtype_list[i]
      code_size = self.code_size_list[i]
      contiguous_size = self.contiguous_size_list[i]
      assert util.check_dtype(data[i], dtype)
      assert data[i].shape[0] == code_size
      data_i = data[i].to(device).reshape(
        code_size // contiguous_size,
        contiguous_size,
        -1,
      ).transpose(1, 2)
      storage = getattr(self, f"_storage_{i}")
      storage[:, address[mask].to(device)] = data_i[:, mask]

  def empty(self):
    super(FlatContainerGroup, self).empty()
    for i in range(self.n_storage):
      storage = getattr(self, f"_storage_{i}")
      storage.fill_(0)
    self._n_items = 0

  def expand(self):
    super(FlatContainerGroup, self).expand()
    n_new = self.expand_step_size
    
    # expand storage
    for i in range(self.n_storage):
      device = self.device_list[i]
      dtype = self.dtype_list[i]
      code_size = self.code_size_list[i]
      contiguous_size = self.contiguous_size_list[i]
      storage = getattr(self, f"_storage_{i}")
      delattr(self, f"_storage_{i}")
      new_storage = torch.zeros(
        code_size // contiguous_size,
        n_new,
        contiguous_size,
        device=device,
        dtype=dtype
      )
      storage = torch.cat([
        storage,
        new_storage
      ], dim=1)
      self.register_buffer(f"_storage_{i}", storage)

  def add(self, data, ids=None, return_address=False):
    assert type(data) in [list, tuple]
    assert len(data) == self.n_storage
    n_data = data[0].shape[1]

    if ids is not None:
      assert util.check_dtype(ids, torch.int64)
      assert ids.shape[0] == n_data
      ids = ids.to(self.device)
    else:
      ids = torch.arange(
        n_data,
        device=self.device,
        dtype=torch.int64
      ) + self.max_id + 1
    
    # expand if necessary
    while (self.capacity - self.n_items) < n_data:
      self.expand()

    # add data to storage
    write_address = torch.arange(
      n_data,
      device=self.device,
      dtype=torch.int64
    ) + self.n_items
    self.set_data_by_address(data=data, address=write_address)
    self._address2id[write_address] = ids

    ### OR
    # for i in range(self.n_storage):
    #   storage = getattr(self, f"_storage_{i}")
    #   device = self.device_list[i]
    #   storage[:, self.n_items:self.n_items+n_data] = data[i].to(device)
    # self._address2id[self.n_items:self.n_items+n_data] = ids
    
    if self.use_inverse_id_mapping:
      self.create_inverse_id_mapping()
    self._n_items += n_data
    
    if return_address:
      return (ids, write_address)
    else:
      return ids

  def remove(self, ids=None, address=None):
    if ids is not None:
      address = self.get_address_by_id(ids)
    elif address is not None:
      util.check_dtype(address, torch.int64)
      address = address.to(self.device)
    else:
      raise RuntimeError("Need either ids or address")

    mask = (0 <= address) & (address < self.capacity)
    address = address[mask].unique(sorted=True)
    n_removed = address.shape[0]
    assert n_removed <= self.n_items
    if n_removed == 0:
      return 0

    self._address2id[address] = -1
    tail_ids = self._address2id[self.n_items-n_removed:self.n_items]
    mask = tail_ids != -1

    for i in range(self.n_storage):
      storage = getattr(self, f"_storage_{i}")
      tail_data = storage[:,self.n_items-n_removed:self.n_items] #
      n = (~mask).sum().item() # number of removed items in tail
      storage[:, address[:n_removed-n]] = tail_data[:, mask].clone()
      storage[:, self.n_items-n_removed:self.n_items] = 0

    self._address2id[address[:n_removed-n]] = tail_ids[mask].clone()
    self._address2id[self.n_items-n_removed:self.n_items] = -1
    if self.use_inverse_id_mapping:
      self.create_inverse_id_mapping()
    self._n_items -= n_removed
    return n_removed


import torch
from ..kernels import TopkSelectCuda

class Topk:
  def __init__(self):
    self._top32_cuda = TopkSelectCuda(
      tpb = 32,
      queue_capacity = 4,
      buffer_size = 4,
    )
    self._top64_cuda = TopkSelectCuda(
      tpb = 64,
      queue_capacity = 4,
      buffer_size = 4,
    )
    self._top128_cuda = TopkSelectCuda(
      tpb = 128,
      queue_capacity = 4,
      buffer_size = 4,
    )
    self._top256_cuda = TopkSelectCuda(
      tpb = 256,
      queue_capacity = 4,
      buffer_size = 4,
    )
    self._top512_cuda = TopkSelectCuda(
      tpb = 512,
      queue_capacity = 2,
      buffer_size = 2,
    )
    self._top1024_cuda = TopkSelectCuda(
      tpb = 1024,
      queue_capacity = 2,
      buffer_size = 2,
    )

  def __call__(self, x, k=1, dim=1):
    if dim == -1:
      dim = 1
    assert dim == 1, "only support last dimention"
    assert len(x.shape) == 2, "only support 2d tensors"
    assert x.is_contiguous(), "x is not contiguous"
    assert k >= 1
    assert x.device.type == "cuda"
    if k == 1:
      return torch.max(x, dim=dim, keepdim=True)
    elif k <= 32:
      return self._top32_cuda(x, dim=dim, k=k)
    elif k <= 64:
      return self._top64_cuda(x, dim=dim, k=k)
    elif k <= 128:
      return self._top128_cuda(x, dim=dim, k=k)
    elif k <= 256:
      return self._top256_cuda(x, dim=dim, k=k)
    elif k <= 512:
      return self._top512_cuda(x, dim=dim, k=k)
    elif k <= 1024:
      return self._top1024_cuda(x, dim=dim, k=k)
    else:
      return torch.topk(x, dim=dim, k=k)



import torch
import torch.nn as nn
import numpy as np
import math
from time import time

from ..kernels import MaxSimCuda
from ..kernels import MinBMMCuda
from ..kernels import TopkBMMCuda
from ..kernels import ComputeCentroidsCuda
from ..CustomModule import CustomModule

class MultiKMeans(CustomModule):
  """
  Run multiple independent K-means algorithms in parallel.
  Parameters:
    n_clusters: int, 
      Number of clusters

    max_iter: int, default: 100
      Maximum number of iterations

    tol: float, default: 0.0001
      Tolerance

    n_redo: int, default: 1
      Number of time k-means will be run with differently intialized centroids.
      the centroids with the lowest inertia will be selected as a final result.

    init_mode: {'random', 'kmeans++'}, default: 'random'
      Initialization method
      'random': randomly chose initial centroids from input data
      'kmeans++': use k-means++ algorithm to initialize centroids, slow when n_cluster is large, but converges faster)
    
    verbose: int, default: 0
      Verbosity

    distance: {'euclidean', 'cosine', 'manhattan'}, default: 'euclidean'
      Type of distance metric
      note: manhattan or L1 distance is only supported on GPU
      
  Attributes:
    centroids: torch.Tensor, shape: [d_vector, n_clusters]
      cluster centroids
  """
  def __init__(
    self,
    n_clusters,
    n_redo=1,
    max_iter=100,
    tol=1e-4,
    distance="euclidean",
    init_mode="random",
    verbose=0,
    sm_size=48*256*4,
   ):
    super(MultiKMeans, self).__init__()
    self.n_redo = n_redo
    self.n_clusters = n_clusters
    self.max_iter = max_iter
    self.tol = tol
    self.verbose = verbose
    self.distance = distance
    self.init_mode = init_mode
    self.sm_size = sm_size
    self.arange = None
    if n_clusters < 4096:
      dk = n_clusters
    else:
      dk = 4096
    de = 1

    self.register_buffer("centroids", None)
    
    if torch.cuda.is_available():
      self.compute_centroids_cuda = ComputeCentroidsCuda(
        de=de,
        dk=dk,
        sm_size=sm_size,
      )

      if distance in ["euclidean"]:
        # self.max_sim_cuda = MinBMMCuda(
        # 4, 4, distance="euclidean",
        # )
        self.max_sim_cuda = MaxSimCuda(
        dim=2, distance="euclidean",
        )
        self.topk_sim_cuda = TopkBMMCuda(
          4, 4, distance="negative_euclidean",
        )
        
      elif distance in ["manhattan"]:
        # self.max_sim_cuda = MinBMMCuda(
        # 4, 4, distance="manhattan",
        # )
        self.max_sim_cuda = MaxSimCuda(
        dim=2, distance="manhattan",
        )
        self.topk_sim_cuda = TopkBMMCuda(
          4, 4, distance="negative_manhattan",
        )

      elif distance in ["cosine"]:
        # self.max_sim_cuda = MinBMMCuda(
        # 4, 4, distance="negative_inner",
        # )
        self.max_sim_cuda = MaxSimCuda(
          dim=2, distance="inner",
        )
        self.topk_sim_cuda = TopkBMMCuda(
          4, 4, distance="inner",
        )
        
      self.warmup_kernels()

  @staticmethod
  def remaining_memory(device):
    """
      Get remaining memory of GPU in bytes
    """
    # torch.cuda.synchronize()
    if device.type == "cpu":
      remaining = 32 * 1024 ** 3 # just a random large number
    elif device.type == "cuda":
      torch.cuda.empty_cache()
      total_memory = torch.cuda.get_device_properties(0).total_memory
      remaining = total_memory - torch.cuda.memory_reserved()
      # remaining = total_memory - torch.cuda.memory_allocated()
    return remaining

  @staticmethod
  def does_it_fit(size, device="cpu", dtype=torch.float):
    try:
      torch.empty(size, device=device, dtype=dtype)
    except:
      return False
    else:
      return True

  @staticmethod
  def calculate_error(a, b):
    """
      Compute L2 error between 'a' and 'b'
    """
    diff = a - b
    diff.pow_(2)
    return diff.sum()

  @staticmethod
  def calculate_inertia(a):
    return (-a).mean()

  @staticmethod
  def cos_sim(a, b, normalize=True, inplace=False):
    """
      Compute batched cosine similarity between 'a' and 'b'

      a: torch.Tensor, shape : [n_kmeans, d_vector, m]
      b: torch.Tensor, shape : [n_kmeans, d_vector, n]
      normalize: bool, default : True
        if True, a and b will be normalized to norm=1
      inplace: bool, default : False
      returns: torch.Tensor, shape : [l, m, n]
    """
    if normalize:
      a_norm = a.norm(dim=-2, keepdim=True) + 1e-8 #[l, m] <l*(m*4)>
      b_norm = b.norm(dim=-2, keepdim=True) + 1e-8 #[l, n] <l*(m*4 + n*4)>
      if inplace:
        # memory consump: m + n + (m * n)
        a.div_(a_norm)
        b.div_(b_norm)
      else:
        # memory consum: m + n + (m * n) + m*d + n*d
        a = a / a_norm #[l, d_vector, m], l*(<m*4 + n*4> + <m*d*4>)
        b = b / b_norm #[l, d_vector, n], l*(<m*4 + n*4> + <(m+n)*d*4>)
    prod = a.transpose(-2, -1) @ b #[l, m, n], <m*n*4 + m*4 + n*4> + <(m+n)*d*4>
    if inplace and normalize:
      a.mul_(a_norm)
      b.mul_(b_norm)
    return prod

  @staticmethod
  def euc_sim(a, b, inplace=False):
    """
      Compute batched negative squared euclidean distance between 'a' and 'b'
      a: torch.Tensor, shape : [l, d_vector, m]
      b: torch.Tensor, shape : [l, d_vector, n]
      inplace: bool, default : False
      returns: torch.Tensor, shape : [l, m, n]
    """
    # peak mem uwage: m*n*4 + max(m,n)*4 + inplace ? 0: (m+n)*d*4
    y = a.transpose(-2, -1) @ b # [m, n] <m*n*4>
    y.mul_(2)
    if inplace:
      a.pow_(2)
      b.pow_(2)
    else:
      a = a ** 2 #[m, d], <m*n*4 + m*d*4>
      b = b ** 2 #[n, d], <m*n*4 + n*d*4 + m*d*4>
    a2 = a.sum(dim=-2)[..., :, None] #? [m], <m*n*4 + m*4> + <n*d*4 + m*d*4>
    y.sub_(a2)
    del a2
    b2 = b.sum(dim=-2)[..., None, :] #[n], <m*n*4 + n*4> + <n*d*4 + m*d*4>
    y.sub_(b2)
    if inplace:
      a.sqrt_()
      b.sqrt_()
    return y
  
  def sim(self, a, b, inplace=False, normalize=True):
    """
      Compute batched similarity between 'a' and 'b', the type of distance metric is specified in __init__ method
      a: torch.Tensor, shape : [l, d, m]
      b: torch.Tensor, shape : [l, d, n]
      returns: torch.Tensor, shape : [l, m, n]
    """
    if self.distance == "euclidean":
      return self.euc_sim(a, b, inplace=inplace)
    elif self.distance == "cosine":
      return self.cos_sim(a, b, inplace=inplace, normalize=normalize)
    elif self.distance == "inner":
      return self.cos_sim(a, b, inplace=inplace, normalize=False)

  def warmup_kernels(self):
    a = torch.randn(128, 128, device="cuda")
    b = torch.randn(128, 128, device="cuda")
    self.max_sim_cuda(a, b, dim=1)
    self.topk_sim_cuda(a, b, dim=1, k=128)

  ### Need more testing:
  def kmeanspp(self, data):
    """
      Initialize centroids with k-means++ algorithm
      data: torch.Tensor, shape : [l, d_vector, n_data]
      returns: torch.Tensor, shape : [l, d_vector, n_clusters]
    """
    l, d_vector, n_data = data.shape
    if self.distance == "cosine":
      data_norm = data.norm(dim=-2, keepdim=True) + 1e-8
      data.div_(data_norm)
    centroids = torch.zeros(l, d_vector, self.n_clusters, device=data.device, dtype=data.dtype)
    #Select initial centroid
    centroids[:, :, 0] = data[:, :, np.random.randint(n_data)]
    for i in range(1, self.n_clusters):
      current_centroids = centroids[:, :, :i].contiguous()
      if data.device.type == "cpu":
        sims = self.sim(data, current_centroids ) #[l,m,n]
        max_sims_v, max_sims_i = sims.max(dim=-1) #[l,m]
      elif data.device.type == "cuda":
        # max_sims_v, max_sims_i = self.max_sim_cuda(
        #   data.transpose(-1, -2),
        #   current_centroids,
        #   dim=2
        # )
        max_sims_v, max_sims_i = self.max_sim_cuda(
          data,
          current_centroids,
          dim=2,
          mode="tn"
        )
      index = max_sims_v.argmin(dim=-1) #[l]
      arange = torch.arange(l, device=data.device)
      new_centroid = data[arange, :, index] #[l, d_vector]
      centroids[:, :, i] = new_centroid
    if self.distance == "cosine":
      data.mul_(data_norm)
    return centroids

  def initialize_centroids(self, data):
    """
      Initialize centroids with init_method specified in __init__
      data: torch.Tensor, shape : [l, d_vector, n_data]
      return: torch.Tensor, shape: [l, d_vector, n_clusters]
    """
    l, d_vector, n_data = data.shape
    if self.init_mode == "random":
      random_index = np.random.choice(
        n_data,
        size=[self.n_clusters],
        replace=False
      )
      centroids = data[:, :, random_index].clone()
      self.print_message("centroids are randomly initialized", 1)

    elif self.init_mode == "kmeans++":
      centroids = self.kmeanspp(data).clone()
      self.print_message("kmeans++ initialization is done!", 1)
    return centroids

  def get_labels(self, data, centroids):
    """
      data: torch.Tensor, shape : [l, d_vector, n_data]
      centroids: torch.Tensor, shape : [l, d_vector, n_clusters]
      return: torch.Tensor, shape : [l, n_data], dtype: long
    """
    #memory requirement: 

    l, d, m = data.shape
    l, d, n = centroids.shape

    remaining = self.remaining_memory(data.device)# - 1024*3

    if self.distance == "euclidean":
      required = l*(m*n + max(m, n) + m*d + n*d) * data.element_size()
    elif self.distance in ["cosine", "inner"]:
      required = l*((m*n) + (m+n)*(d+1)) * data.element_size()
    # if remaining >= required:
    if False:
      sims = self.sim(data, centroids, inplace=False) #[l, m, n]
      maxsims, labels = sims.max(dim=-1) #[l, m]
      return (maxsims, labels)
    else:
      if data.device.type == "cuda":
        if self.distance == "cosine":
          d_norm = data.norm(dim=-2, keepdim=True) + 1e-8
          c_norm = centroids.norm(dim=-2, keepdim=True) + 1e-8
          data.div_(d_norm)
          centroids.div_(c_norm)
        # maxsims, labels = self.max_sim_cuda(
        #   data.transpose(-1, -2),
        #   centroids,
        #   dim=2
        # )
        maxsims, labels = self.max_sim_cuda(
          data,
          centroids,
          dim=2,
          mode="tn"
        )
        if self.distance == "cosine":
          data.mul_(d_norm)
          centroids.mul_(c_norm)
      elif data.device.type == "cpu":
        ## doing in seperate chunks
        n_partitions = 1
        for i in range(16):
          sub_m = math.ceil(m / n_partitions)
          if self.distance == "euclidean":
            required = l*(sub_m*n + max(sub_m, n)) * data.element_size() + m*8 # +sub_m*d*4
          elif self.distance in ["cosine", "inner"]:
            required = l*(sub_m*n + sub_m+n) * data.element_size() + m*8# +sub_m*d*4
          # print("required, remaining, n_p", required / 1024**3, remaining / 1024**3, n_partitions)
          if self.does_it_fit(required // 4, device=data.device, dtype=torch.float):
            break
          n_partitions *= 2

        maxsims = torch.zeros(l, m, device=data.device, dtype=torch.float)
        labels = torch.zeros(l, m, device=data.device, dtype=torch.long)
        for i in range(n_partitions):
          start = i*sub_m
          if i == n_partitions - 1:
            end = m - 1
          else:
            end = (i+1)*sub_m
          sub_data = torch.narrow(data, dim=-1, start=start, length=end-start) #[l, d, sub_m]
          # sub_data = data[:, start:end] #[l, d, sub_m]
          sub_sims = self.sim(sub_data, centroids, inplace=True) #[l, sub_m, n]
          del sub_data
          sub_maxsims, sub_labels = sub_sims.max(dim=-1) #[l, sub_m]
          del sub_sims
          labels[:, start:end] = sub_labels
          maxsims[:, start:end] = sub_maxsims
          del sub_labels
      return (maxsims, labels)

  def compute_centroids_loop(self, data, labels):
    """
      data: torch.Tensor, shape : [l, d_vector, n_data]
      labels: torch.Tensor, shape : [l, n_data]
      return: torch.Tensor, shape : [l, d_vector, n_clusters]
    """
    ### Naive method with loop
    l, d, m = data.shape
    centroids = torch.zeros(l, d, self.n_clusters, device=data.device, dtype=data.dtype)
    for j in range(l):
      unique_labels, counts = labels[j].unique(return_counts=True)
      for i, count in zip(unique_labels, counts):
        centroids[j, :, i] = data[j, :, labels[j]==i].sum(dim=1) / count
    return centroids

  def compute_centroids(self, data, labels):
    """
      data: torch.Tensor, shape : [l, d_vector, n_data]
      labels: torch.Tensor, shape : [l, n_data]
      return: torch.Tensor, shape: [l, d_vector, n_clusters]
    """
    if data.device == torch.device("cpu"):
      centroids = self.compute_centroids_loop(data, labels)
    else:
      centroids = self.compute_centroids_cuda(data, labels, k=self.n_clusters)
    return centroids

  def _compute_centroids_hungry(self, data, labels):
    ### Memory hungry method
    # expanded_labels = labels[None].expand(self.n_clusters, -1) #[k, n], k=n_clusters <>
    if self.arange is None\
    or self.arange.dtype != data.dtype\
    or self.arange.device != data.device:
      self.arange = torch.arange(self.n_clusters, device=data.device) #[k] <k*8>
    
    mask = labels[None, :] == self.arange[:, None] #[k, n] <k*n*1 + k*8>
    mask_sum = mask.sum(dim=-1) #[k] <k*n*1 + k*12>
    mask = mask.float() # <k*n*5 + k*12> LARGEST MEMORY USE!!!
    centroids = mask @ data # <k*n*4 + k*12 + k*d*4>
    del mask

    centroids.div_(mask_sum[..., :, None]) # <k*d*4 + k*12>
    del mask_sum
    nan_mask = centroids!=centroids #[k, d] # <k*d*8>
    centroids[nan_mask] = 0 # remove NaNs

    return centroids

  def fit(self, data, centroids=None):
    """
      Perform K-means clustering, and return final labels
      data: torch.Tensor, shape : [l, d_vector, n_data]
      return: torch.Tensor, shape : [l, n_data], dtype: long
    """
    assert data.is_contiguous(), "use .contiguous()"

    best_centroids = None
    best_error = 1e32
    best_labels = None
    best_inertia = 1e32
    tm = time()
    for i in range(self.n_redo):
      tm_i = time()
      if centroids is None:
        centroids = self.initialize_centroids(data)
      for j in range(self.max_iter):
        # 1 iteration of clustering
        maxsims, labels = self.get_labels(data, centroids) #top1 search
        new_centroids = self.compute_centroids(data, labels)
        error = self.calculate_error(centroids, new_centroids)
        centroids = new_centroids
        inertia = self.calculate_inertia(maxsims)
        self.print_message(f"----iteration {j} of {i}th redo, error={error.item()}, inertia={inertia.item()}", 3)
        if error <= self.tol:
          break

      if inertia < best_inertia:
        best_centroids = centroids
        best_error = error
        best_labels = labels
        best_inertia = inertia
      self.print_message(f"--{i}th redo finished, error: {error.item()}, inertia: {inertia.item()}time spent:{round(time()-tm_i, 4)} sec", 2)

    self.register_buffer("centroids", best_centroids)
    self.print_message(f"finished {self.n_redo} redos in {round(time()-tm, 4)} sec, final_inertia: {best_inertia}", 1)
    return best_labels

  def predict(self, query):
    """
      Predict closest cluster center each sample in query belongs to.
      query: torch.Tensor, shape : [l, d_vector, n_query]
      return: torch.Tensor, shape : [l, n_query]
    """
    assert self.centroids is not None, "kmeans is not trained"
    _, labels = self.get_labels(query, self.centroids)
    return labels

  def topk(self, query, k=128):
    """
      Predict the top-k closest cluster centers of each sample in query
      query: torch.Tensor, shape : [l, d_vector, n_query]
      k: int, should be in range [1, n_centroids]
    """
    assert self.centroids is not None, "kmeans is not trained"
    assert k <= self.n_clusters, "k is larger than number of clusters"
    if k == 1:
      # topk_v, topk_i = self.max_sim_cuda(
      #   query.transpose(-1, -2),
      #   self.centroids,
      #   dim=2
      # )
      topk_v, topk_i = self.max_sim_cuda(
        query,
        self.centroids,
        dim=2,
        mode="tn"
      )
      return (topk_v[..., None], topk_i[..., None])
    elif k <= 128:
      return self.topk_sim_cuda(
        query.transpose(-1, -2),
        self.centroids,
        dim=2,
        k=k
      )
    elif k > 128:
      sims = self.sim(query, self.centroids) #[n_query, n_clusters]
      topk_v, topk_i = sims.topk(dim=-1, k=k) #[n_query, k]
      return (topk_v, topk_i)



import torch

from ..clustering import KMeans
# from ..clustering import KMeansOld as KMeans
from .BaseCodec import BaseCodec

class VQCodec(BaseCodec):
  def __init__(self, *args, **kwargs):
    super(VQCodec, self).__init__()
    self.kmeans = KMeans(
      *args,
      **kwargs
    )

  @property
  def codebook(self):
    return self.kmeans.centroids

  def encode(self, input):
    """
      input:
        torch.Tensor
        shape : [d_vector, n_data]
        dtype : float32

      returns:
        torch.Tensor
        shape : [n_data]
        dtype : int64
    """
    assert self._is_trained == True, "codec is not trained"
    return self.kmeans.predict(input)

  def decode(self, code):
    """
      code:
        torch.Tensor
        shape : [n_code]
        dtype : int64

      returns:
        torch.Tensor
        shape : [d_vector, n_data]
        dtype : float32
    """
    assert self._is_trained == True, "Codec is untrained"
    return self.codebook[:, code].clone()

  def train(self, data):
    """
      data:
        torch.Tensor
        shape : [d_vector, n_data]
        dtype : float32
    """
    labels = self.kmeans.fit(data)
    self._trained(True)
    return labels


try:
  import cupy as cp
except ModuleNotFoundError:
  raise ModuleNotFoundError("cupy is not installed, please visit https://pypi.org/project/cupy/")

from . import container
from . import codec
from . import transform
from . import clustering
from . import kernels
from . import experimental
from . import metric
from . import util
from . import legacy
from . import index
from . import fn

from .CustomModule import CustomModule

topk = fn.Topk()



import torch
import numpy as np
from .BaseIndex import BaseIndex
from ..container import CellContainer
from ..codec import PQCodec, VQCodec
from ..fn import IVFPQTopk
from ..fn import Topk
from .. import util
from .. import metric

class IVFPQIndex(CellContainer):
  def __init__(
      self,
      d_vector,
      n_subvectors=8,
      n_cells=128,
      initial_size=None,
      expand_step_size=128,
      expand_mode="double",
      distance="euclidean",
      device='cuda:0',
      pq_use_residual=False,
      verbose=0,
    ):
    if torch.device(device).type == "cuda":
      assert torch.cuda.is_available(), "cuda is not available"
      max_sm_bytes = util.get_maximum_shared_memory_bytes()
      assert n_subvectors <= max_sm_bytes // 1024
    assert d_vector % n_subvectors == 0

    super(IVFPQIndex, self).__init__(
      code_size = n_subvectors,
      n_cells = n_cells,
      dtype = "uint8",
      device = device,
      initial_size = initial_size,
      expand_step_size = expand_step_size,
      expand_mode = expand_mode,
      use_inverse_id_mapping = True,
      contiguous_size = 4,
      verbose = verbose,
    )

    self.d_vector = d_vector
    self.n_subvectors = n_subvectors
    self.d_subvector = d_vector // n_subvectors
    self.distance = distance
    self.verbose = verbose
    self.pq_use_residual = pq_use_residual
    self.n_probe = 1
    if pq_use_residual and (n_cells * 256 * n_subvectors * 4) <= 4*1024**3:
      self._use_precomputed = True
    else:
      self._use_precomputed = False
    self._precomputed_part2 = None
    self._use_cublas = True
    self._use_smart_probing = True
    self._smart_probing_temperature = 30.0

    self.vq_codec = VQCodec(
      n_clusters = n_cells,
      n_redo = 1,
      max_iter = 15,
      tol = 1e-4,
      distance = "euclidean",
      init_mode = "random",
      verbose = verbose
    )

    self.pq_codec = PQCodec(
      d_vector = d_vector,
      n_subvectors = n_subvectors,
      n_clusters = 256,
      distance = distance,
      verbose = verbose
    )

    self._ivfpq_topk = IVFPQTopk(
      n_subvectors = n_subvectors,
      contiguous_size = self.contiguous_size,
      sm_size = n_subvectors * 1024,
    )

    self._topk = Topk()

  @property
  def use_cublas(self):
    return self._use_cublas

  @use_cublas.setter
  def use_cublas(self, value):
    assert type(value) is bool
    self._use_cublas = value

  @property
  def use_smart_probing(self):
    return self._use_smart_probing

  @use_smart_probing.setter
  def use_smart_probing(self, value):
    assert type(value) is bool
    self._use_smart_probing = value

  @property
  def smart_probing_temperature(self):
    return self._smart_probing_temperature

  @smart_probing_temperature.setter
  def smart_probing_temperature(self, value):
    assert value > 0
    assert self.use_smart_probing, "set use_smart_probing to True first"
    self._smart_probing_temperature = value

  @property
  def use_precomputed(self):
    return self._use_precomputed

  @use_precomputed.setter
  def use_precomputed(self, value):
    assert type(value) is bool
    if value:
      assert self.pq_use_residual, " `use_precomputed=True` is only valid when `pq_use_residual` is True"
      assert (self.pq_codec.is_trained and self.vq_codec.is_trained), "index is not trained"
      required_memory = (self.n_cells * self.n_subvectors * 1024) / 1024**2
      self.print_message(f"estimated memory for precomputed table is {required_memory} MB", 1)
      self.precompute_part2()
    else:
      self._precomputed_part2 = None
    self._use_precomputed = value

  def precompute_part2(self):
    pq_codebook = self.pq_codec.codebook
    vq_codebook = self.vq_codec.codebook.reshape(
      self.n_subvectors,
      self.d_subvector,
      self.n_cells
    )
    self._precomputed_part2 = torch.bmm(
      vq_codebook.transpose(-1, -2),
      pq_codebook
    ) * -2 - pq_codebook.norm(dim=1).pow(2)[:, None]

  @property
  def vq_codec_max_iter(self):
    return self.vq_codec.kmeans.max_iter

  @vq_codec_max_iter.setter
  def vq_codec_max_iter(self, value):
    assert type(value) is int
    assert value > 0
    assert not self.vq_codec.is_trained, "vq_codec is already trained"
    self.vq_codec.kmeans.max_iter = value

  @property
  def vq_codec_n_redo(self):
    return self.vq_codec.kmeans.n_redo

  @vq_codec_n_redo.setter
  def vq_codec_n_redo(self, value):
    assert type(value) is int
    assert value > 0
    assert not self.vq_codec.is_trained, "vq_codec is already trained"
    self.vq_codec.kmeans.n_redo = value
  
  @property
  def vq_codec_tolerance(self):
    return self.vq_codec.kmeans.tol

  @vq_codec_tolerance.setter
  def vq_codec_tolerance(self, value):
    assert not self.vq_codec.is_trained, "vq_codec is already trained"
    self.vq_codec.kmeans.tol = value

  @property
  def pq_codec_max_iter(self):
    return self.pq_codec.kmeans.max_iter
  
  @pq_codec_max_iter.setter
  def pq_codec_max_iter(self, value):
    assert type(value) is int
    assert value > 0
    assert not self.pq_codec.is_trained, "pq_codec is already trained"
    self.pq_codec.kmeans.max_iter = value

  @property
  def pq_codec_n_redo(self):
    return self.pq_codec.kmeans.n_redo

  @pq_codec_n_redo.setter
  def pq_codec_n_redo(self, value):
    assert type(value) is int
    assert value > 0
    assert not self.pq_codec.is_trained, "pq_codec is already trained"
    self.pq_codec.kmeans.n_redo = value

  @property
  def pq_codec_tolerance(self):
    return self.pq_codec.kmeans.tol
  
  @pq_codec_tolerance.setter
  def pq_codec_tolerance(self, value):
    assert not self.pq_codec.is_trained, "pq_codec is already trained"
    self.pq_codec.kmeans.tol = value

  def train(self, x, force_retrain = False):
    if self.vq_codec.is_trained and self.pq_codec.is_trained:
      if not force_retrain:
        self.print_message("index is already trained", 1)
        return
    assert len(x.shape) == 2
    assert x.shape[0] == self.d_vector
    if self.distance == "cosine":
      x = util.normalize(x, dim=0)
    d_vector, n_data = x.shape

    self.print_message("start training VQ codec...", 1)
    code = self.vq_codec.train(x)

    self.print_message("start training PQ codec...", 1)
    if self.pq_use_residual:
      recon = self.vq_codec.decode(code)
      x.sub_(recon)
      del recon
      self.pq_codec.train(x)
      recon = self.vq_codec.decode(code)
      x.add_(recon)
      del recon
    else:
      self.pq_codec.train(x)

    self.print_message("index is trained successfully!", 1)
  
  def encode(self, x):
    """
      Encode `x` with PQ codec
      x:
        torch.Tensor
        dtype : float32
        shape : [d_vector, n_data]

      returns:
        torch.Tensor
        dtype : uint8
        shape : [n_subvectors, n_data]
    """
    assert len(x.shape) == 2
    assert x.shape[0] == self.d_vector
    if self.distance == "cosine":
      x = util.normalize(x)
    if self.pq_use_residual:
      vq_code = self.vq_codec.encode(x)
      recon = self.vq_codec.decode(vq_code)
      pq_code = self.pq_codec.encode(x - recon)
      return pq_code, vq_code
    else:
      y = self.pq_codec.encode(x)
      return y

  def decode(self, x):
    """
      Decode `x` with PQ codec
      x:
        torch.Tensor
        dtype : uint8
        shape : [n_subvectors, n_data]

      returns:
        torch.Tensor
        dtype : float32
        shape : [d_vector, n_data]
    """
    if self.pq_use_residual:
      assert len(x) == 2
      pq_code, vq_code = x
      assert pq_code.shape[0] == self.n_subvectors
      assert pq_code.shape[1] == vq_code.shape[0]
      residual = self.pq_codec.decode(pq_code)
      recon = self.vq_codec.decode(vq_code)
      y = recon + residual

    else:
      assert len(x.shape) == 2
      assert x.shape[0] == self.n_subvectors
      y = self.pq_codec.decode(x)
    return y
  
  def add(self, x, ids=None, return_address=False):
    """
      Add `x` to index, with optional `ids` for each vector in `x`
      x:
        torch.Tensor
        dtype : float32
        shape : [d_vector, n_data]

      ids: optional
        torch.Tensor
        dtype : int64
        shape : [n_data]
        If not given, or given None, `ids` will be set to 
        torch.arange(n_data) + self.max_id + 1

      return_address:
        bool
        default : False
        if set to True, return address of the added vectors

      returns (ids) or (ids, address):
        ids:
          torch.Tensor
          dtype : int64
          shape : [n_data]
        
        address:
          torch.Tensor
          dtype : int64
          shape : [n_data]
          this is returned if `return_address` is True
    """
    assert len(x.shape) == 2
    assert x.shape[0] == self.d_vector
    if self.distance == "cosine":
      x = util.normalize(x)

    assigned_cells = self.vq_codec.encode(x)
    if self.pq_use_residual:
      quantized_x, _ = self.encode(x)
    else:
      quantized_x = self.encode(x)

    return super(IVFPQIndex, self).add(
      quantized_x,
      cells=assigned_cells,
      ids=ids,
      return_address = return_address
    )

  def precomputed_adc_residual_precomputed(self, x):
    n_query = x.shape[1]
    vq_codebook = self.vq_codec.codebook #[d_vector, n_cells]
    pq_codebook = self.pq_codec.codebook #[n_subvectors, d_subvectors, 256]
    x = x.reshape(
      self.n_subvectors,
      self.d_subvector,
      n_query
    ).transpose(-1, -2)
    part1 = 2 * (x @ pq_codebook).permute(1, 0, 2) #[n_query, n_sub, 256]
    if self._precomputed_part2 is None:
      self.precompute_part2()
    part2 = self._precomputed_part2.transpose(0, 1)
    return part1, part2

  def precomputed_adc_residual(self, x, cells):
    # shape of cells: [n_query, n_probe]
    # shape of x: [d_vector, n_query]
    n_query = x.shape[1]
    n_probe = cells.shape[1]
    vq_codebook = self.vq_codec.codebook #[d_vector, n_cells]
    pq_codebook = self.pq_codec.codebook #[n_subvectors, d_subvectors, 256]
    x = x.reshape(
      self.n_subvectors,
      self.d_subvector,
      n_query
    ).transpose(-1, -2)
    part1 = 2 * (x @ pq_codebook).permute(1, 0, 2) - pq_codebook.norm(dim=1).pow(2)[None]#[n_query, n_sub, 256]

    vq_codebook = vq_codebook.reshape(
      self.n_subvectors,
      self.d_subvector,
      self.n_cells
    )
    vq_codebook = vq_codebook[:, :, cells]#[n_sub, d_sub, n_query, n_probe]
    vq_codebook = vq_codebook.permute(0, 2, 3, 1)#[n_sub, n_query, n_probe, d_sub]
    pq_codebook = pq_codebook[:, None].expand(-1, n_query, -1, -1)#[n_sub, n_query, d_sub, 256]
    part2 = - 2 * (vq_codebook @ pq_codebook)#[n_sub, n_query, n_probe, 256]
    precomputed = part1[:, None] + part2.permute(1, 2, 0, 3)#[n_query, n_probe, n_sub, 256]
    return precomputed

  def search_cells(
      self,
      x,
      cells,
      base_sims=None,
      n_probe_list=None,
      k=1,
      return_address=False
    ):
    storage = self._storage
    is_empty = self._is_empty
    vq_codebook = self.vq_codec.codebook

    cell_start = self._cell_start[cells]
    cell_size = self._cell_size[cells]
    
    if self.pq_use_residual:
      assert base_sims is not None, "base_sims is required when pq_use_residual is True"
      if self.use_precomputed:
        part1, part2 = self.precomputed_adc_residual_precomputed(x)
        topk_val, topk_address = self._ivfpq_topk.topk_residual_precomputed(
          data=storage,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          k=k
        )
      else:
        precomputed = self.precomputed_adc_residual(x, cells)
        topk_val, topk_address = self._ivfpq_topk.topk_residual(
          data=storage,
          base_sims=base_sims,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          k=k
        )
    else:
      precomputed = self.pq_codec.precompute_adc(x)
      topk_val, topk_address = self._ivfpq_topk.topk(
        data=storage,
        precomputed=precomputed,
        cell_start=cell_start,
        cell_size=cell_size,
        is_empty=is_empty,
        n_probe_list=n_probe_list,
        k=k
      )

    topk_ids = self.get_id_by_address(topk_address)
    if return_address:
      return topk_val, topk_ids, topk_address
    else:
      return topk_val, topk_ids

  def search(self, x, k=1, return_address=False):
    assert len(x.shape) == 2
    assert x.shape[0] == self.d_vector
    assert 0 < k <= 1024
    if self.distance == "cosine":
      x = util.normalize(x, dim=0)
    n_query = x.shape[1]
    storage = self._storage
    is_empty = self._is_empty
    vq_codebook = self.vq_codec.codebook

    # find n_probe closest cells
    if self.use_cublas:
      # sims = metric.negative_squared_l2_distance(x.half(), vq_codebook.half())
      sims = metric.negative_squared_l2_distance(x, vq_codebook, use_tensor_core=False)
      sims = sims.float().contiguous()
      topk_sims, cells = self._topk(sims, k=self.n_probe, dim=1)
    else:
      topk_sims, cells = self.vq_codec.kmeans.topk(x, k=self.n_probe)
    
    if self.use_smart_probing and self.n_probe > 1:
      p = -topk_sims.abs().sqrt()
      p = torch.softmax(p / self.smart_probing_temperature, dim=-1)

      # p_norm = p.norm(dim=-1)
      # sqrt_d = self.n_probe ** 0.5
      # score = 1 - (p_norm * sqrt_d - 1) / (sqrt_d - 1) - 1e-6
      # n_probe_list = torch.ceil(score * (self.n_probe) ).long()
      
      max_n_probe = torch.tensor(self.n_probe, device=self.device)
      normalized_entropy = - torch.sum(p * torch.log2(p) / torch.log2(max_n_probe), dim=-1)
      n_probe_list = torch.ceil(normalized_entropy * max_n_probe ).long()
    else:
      n_probe_list = None

    return self.search_cells(
      x=x, 
      cells=cells,
      base_sims=topk_sims,
      n_probe_list=n_probe_list,
      k=k,
      return_address=False
    )
    
    


from .CustomKernel import CustomKernel
from .CustomKernel import Stream

from .GetAddressByIDCuda import GetAddressByIDCuda
from .GetDivByAddressCuda import GetDivByAddressCuda
from .GetDivByAddressV2Cuda import GetDivByAddressV2Cuda
from .GetIOACuda import GetIOACuda
from .GetWriteAddressCuda import GetWriteAddressCuda
from .GetWriteAddressV2Cuda import GetWriteAddressV2Cuda
from .PQDecodeCuda import PQDecodeCuda
from .ComputeCentroidsCuda import ComputeCentroidsCuda
from .ComputeProductCuda import ComputeProductCuda
from .MaxSimCuda import MaxSimCuda

from .BMMCuda import BMMCuda
from .MinBMMCuda import MinBMMCuda
from .TopkBMMCuda import TopkBMMCuda

from .IVFPQTopkCuda import IVFPQTopkCuda
from .IVFPQTop1Cuda import IVFPQTop1Cuda
from .TopkSelectCuda import TopkSelectCuda


import torch
import cupy as cp
import numpy as np
import math

from .CustomKernel import CustomKernel
from ..util import get_absolute_path

class TopkSelectCuda(CustomKernel):
  """
    tpb: threads per block, needs to be a power of 2 between 32 and 1024
    queue_capacity: capacity of thread queue
    buffer_size: number of elements each threads needs to prefetch
  """
  def __init__(self, tpb=256, queue_capacity=4, buffer_size=4):
    super().__init__()
    assert tpb >= 32
    assert self.next_power_of_2(tpb) == tpb
    assert queue_capacity >= 1
    assert buffer_size >= 1
    self.tpb = tpb
    self.queue_capacity = queue_capacity
    self.buffer_size = buffer_size

    with open(get_absolute_path("kernels", "cuda", "topk_select.cu"),'r') as f: ###
      self.kernel = f.read()
    
    self.kernel = (
      self.kernel
      .replace("_TPB_", str(tpb))
      .replace("_QCAP_", str(queue_capacity))
      .replace("_TN_", str(buffer_size))
    )

    self._fn = cp.RawKernel(
      code=self.kernel,
      name="topk_select",
      backend='nvcc',
      options=(
        '--use_fast_math',
        # '--maxrregcount=128',
        #'-Xptxas',
        #'-dlcm=cg',
      )
    )

  @staticmethod
  def next_power_of_2(x):
    return 1 if x == 0 else 2**math.ceil(math.log2(x))
  
  def __call__(self, x, k=128, dim=1):
    """
      x: shape = [m, n], dtype: float32
      k: 1 to 1024
      dim: 1
    """
    assert len(x.shape) == 2
    assert x.dtype in [torch.float32]
    assert x.device.type == "cuda"
    assert 1 <= k <= self.tpb
    assert dim == 1
    assert x.is_contiguous()
    k_pow_of_2 = self.next_power_of_2(k)

    m, n = x.shape
    threads_per_block = (self.tpb, )
    blocks_per_grid = (m, )
    values = torch.empty(m, k_pow_of_2, device="cuda:0", dtype=x.dtype)
    values.fill_(float("-inf"))
    indices = torch.empty(m, k_pow_of_2, device="cuda:0", dtype=torch.long)

    self._fn(
      grid = blocks_per_grid,
      block = threads_per_block,
      args = [
        x.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        m, n, k_pow_of_2
      ],
      stream=self.stream
    )
    return values[:, :k], indices[:, :k]


import torch
import torch.nn as nn
import numpy as np
import math
from time import time

from ..kernels import MaxSimCuda
from ..kernels import MinBMMCuda
from ..kernels import TopkBMMCuda
from ..kernels import ComputeCentroidsCuda
from ..CustomModule import CustomModule

class MinibatchKMeans(CustomModule):
  """
    Minibatch K-means clustering algorithm implemented with pytorch and CUDA
    Parameters:
      n_clusters: int, 
        Number of clusters

      init_mode: {'random', 'kmeans++'}, default: 'random'
        Initialization method
        'random': randomly chose initial centroids from input data
        'kmeans++': use k-means++ algorithm to initialize centroids, slow when n_cluster is large, but converges faster)
      
      verbose: int, default: 0
        Verbosity

      distance: {'euclidean', 'cosine', 'manhattan'}, default: 'euclidean'
        Type of distance metric
        note: manhattan (L1) distance is only supported on GPU
        
    Attributes:
      centroids: torch.Tensor, shape: [d_vector, n_clusters]
        cluster centroids

    Example:
      1| from torchpq.clustering import MinibatchKMeans
      2| minibatch_kmeans = MinibatchKMeans(n_clusters = 128)
      3| n_iter = 10
      4| tol = 0.001
      5| for i in range(n_iter):
      6|   x = next(data_generator)
      7|   minibatch_kmeans.fit_minibatch(x)
      8|   if minibatch_kmeans.error < tol:
      9|     break
  """
  def __init__(
    self,
    n_clusters,
    distance="euclidean",
    init_mode="random",
    verbose=0,
    sm_size=48*256*4,
   ):
  
    super(MinibatchKMeans, self).__init__()
    self.n_clusters = n_clusters
    self.verbose = verbose
    self.distance = distance
    self.init_mode = init_mode
    self.sm_size = sm_size
    self.arange = None
    self._iteration = 0
    self._inertia = 1e32
    self._error = 1e32
    if n_clusters < 4096:
      dk = n_clusters
    else:
      dk = 4096
    de = 1

    self.register_buffer("centroids", None)
    self.register_buffer("n_points_in_clusters", None)
    
    if torch.cuda.is_available():
      self.compute_centroids_cuda = ComputeCentroidsCuda(
        de=de,
        dk=dk,
        sm_size=sm_size,
      )

      if distance in ["euclidean"]:
        # self.max_sim_cuda = MinBMMCuda(
        # 4, 4, distance="euclidean",
        # )
        self.max_sim_cuda = MaxSimCuda(
        dim=2, distance="euclidean",
        )
        self.topk_sim_cuda = TopkBMMCuda(
          4, 4, distance="negative_euclidean",
        )
        
      elif distance in ["manhattan"]:
        # self.max_sim_cuda = MinBMMCuda(
        # 4, 4, distance="manhattan",
        # )
        self.max_sim_cuda = MaxSimCuda(
        dim=2, distance="manhattan",
        )
        self.topk_sim_cuda = TopkBMMCuda(
          4, 4, distance="negative_manhattan",
        )

      elif distance in ["cosine"]:
        # self.max_sim_cuda = MinBMMCuda(
        # 4, 4, distance="negative_inner",
        # )
        self.max_sim_cuda = MaxSimCuda(
          dim=2, distance="inner",
        )
        self.topk_sim_cuda = TopkBMMCuda(
          4, 4, distance="inner",
        )
        
      self.warmup_kernels()

  @property
  def inertia(self):
    return self._inertia

  @property
  def error(self):
    return self._error

  @staticmethod
  def remaining_memory(device):
    """
      Get remaining memory of GPU in bytes
    """
    # torch.cuda.synchronize()
    if device.type == "cpu":
      remaining = 32 * 1024 ** 3 # just a random large number
    elif device.type == "cuda":
      torch.cuda.empty_cache()
      total_memory = torch.cuda.get_device_properties(0).total_memory
      remaining = total_memory - torch.cuda.memory_reserved()
      # remaining = total_memory - torch.cuda.memory_allocated()
    return remaining

  @staticmethod
  def does_it_fit(size, device="cpu", dtype=torch.float):
    try:
      torch.empty(size, device=device, dtype=dtype)
    except:
      return False
    else:
      return True

  @staticmethod
  def calculate_error(a, b):
    diff = a - b
    diff.pow_(2)
    return diff.sum()

  @staticmethod
  def calculate_inertia(a):
    return (-a).mean()

  @staticmethod
  def cos_sim(a, b, normalize=True, inplace=False):
    """
      Computes cosine similarity between 'a' and 'b'

      a: torch.Tensor, shape : [d_vector, m]
      b: torch.Tensor, shape : [d_vector, n]
      normalize: bool, default : True
        if True, a and b will be normalized to norm=1
      inplace: bool, default : False
      return: torch.Tensor, shape : [m, n]
    """
    if normalize:
      a_norm = a.norm(dim=0, keepdim=True) + 1e-8 #[m] <m*4>
      b_norm = b.norm(dim=0, keepdim=True) + 1e-8 #[n] <m*4 + n*4>
      if inplace:
        # memory consump: m + n + (m * n)
        a.div_(a_norm)
        b.div_(b_norm)
      else:
        # memory consum: m + n + (m * n) + m*d + n*d
        a = a / a_norm #[d_vector, m], <m*4 + n*4> + <m*d*4>
        b = b / b_norm #[d_vector, n], <m*4 + n*4> + <(m+n)*d*4>
    prod = a.transpose(-2, -1) @ b #[m, n], <m*n*4 + m*4 + n*4> + <(m+n)*d*4>
    if inplace and normalize:
      a.mul_(a_norm)
      b.mul_(b_norm)
    return prod

  @staticmethod
  def euc_sim(a, b, inplace=False):
    """
      Computes negative squared euclidean distance between 'a' and 'b'

      a: torch.Tensor, shape : [d_vector, m]
      b: torch.Tensor, shape : [d_vector, n]
      inplace: bool, default : False
      return: torch.Tensor, shape : [m, n]
    """
    # peak mem uwage: m*n*4 + max(m,n)*4 + inplace ? 0: (m+n)*d*4
    y = a.transpose(-2, -1) @ b # [m, n] <m*n*4>
    y.mul_(2)
    if inplace:
      a.pow_(2)
      b.pow_(2)
    else:
      a = a ** 2 #[m, d], <m*n*4 + m*d*4>
      b = b ** 2 #[n, d], <m*n*4 + n*d*4 + m*d*4>
    a2 = a.sum(dim=0)[..., :, None] #? [m], <m*n*4 + m*4> + <n*d*4 + m*d*4>
    y.sub_(a2)
    del a2
    b2 = b.sum(dim=0)[..., None, :] #[n], <m*n*4 + n*4> + <n*d*4 + m*d*4>
    y.sub_(b2)
    if inplace:
      a.sqrt_()
      b.sqrt_()
    return y
  
  def sim(self, a, b, inplace=False, normalize=True):
    """
      Computes similarity between 'a' and 'b'
      a: torch.Tensor, shape : [d, m]
      b: torch.Tensor, shape : [d, n]
      returns: torch.Tensor, shape : [m, n]
    """
    if self.distance == "euclidean":
      return self.euc_sim(a, b, inplace=inplace)
    elif self.distance == "cosine":
      return self.cos_sim(a, b, inplace=inplace, normalize=normalize)
    elif self.distance == "inner":
      return self.cos_sim(a, b, inplace=inplace, normalize=False)

  def warmup_kernels(self):
    a = torch.randn(128, 128, device="cuda")
    b = torch.randn(128, 128, device="cuda")
    self.max_sim_cuda(a, b, dim=1)
    self.topk_sim_cuda(a, b, dim=1, k=128)

  def kmeanspp(self, data):
    """
      K-means++ initialization

      data: torch.Tensor, shape : [d_vector, n_data]
      returns: torch.Tensor, shape : [d_vector, n_clusters]
    """
    d_vector, n_data = data.shape
    if self.distance == "cosine":
      data_norm = data.norm(dim=0, keepdim=True) + 1e-8
      data.div_(data_norm)
    centroids = torch.zeros(d_vector, self.n_clusters, device=data.device, dtype=data.dtype)
    #Select initial centroid
    centroids[:, 0] = data[:, np.random.randint(n_data)]
    for i in range(1, self.n_clusters):
      current_centroids = centroids[:, :i].contiguous()
      if data.device.type == "cpu":
        sims = self.sim(data, current_centroids )
        max_sims_v, max_sims_i = sims.max(dim=1)
      elif data.device.type == "cuda":
        # max_sims_v, max_sims_i = self.max_sim_cuda(data.transpose(-1, -2), current_centroids, dim=1)
        max_sims_v, max_sims_i = self.max_sim_cuda(
          data, 
          current_centroids, 
          dim=1, 
          mode="tn"
        )
      index = max_sims_v.argmin(dim=0)
      new_centroid = data[:, index]
      centroids[:, i] = new_centroid
    if self.distance == "cosine":
      data.mul_(data_norm)
    return centroids

  def initialize_centroids(self, data):
    """
      Initializes centroids
      data: torch.Tensor, shape : [d_vector, n_data]
      return: torch.Tensor, shape: [d_vector, n_clusters]
    """
    d_vector, n_data = data.shape
    if self.init_mode == "random":
      random_index = np.random.choice(
        n_data,
        size=[self.n_clusters],
        replace=False
      )
      centroids = data[:, random_index].clone()
      self.print_message("centroids are randomly initialized", 1)

    elif self.init_mode == "kmeans++":
      centroids = self.kmeanspp(data).clone()
      self.print_message("kmeans++ initialization is done!", 1)
    return centroids

  def get_labels(self, data, centroids):
    """
      data: torch.Tensor, shape : [d_vector, n_data]
      centroids: torch.Tensor: shape : [d_vector, n_clusters]
    """
    #memory requirement: 

    d, m = data.shape
    d, n = centroids.shape

    remaining = self.remaining_memory(data.device)# - 1024*3

    if self.distance == "euclidean":
      required = (m*n + max(m, n) + m*d + n*d) * data.element_size()
    elif self.distance in ["cosine", "inner"]:
      required = ((m*n) + (m+n)*(d+1)) * data.element_size()
    # if remaining >= required:
    if False:
      sims = self.sim(data, centroids, inplace=False) #[m, n]
      maxsims, labels = sims.max(dim=-1) #[m]
      return (maxsims, labels)
    else:
      if data.device.type == "cuda":
        if self.distance == "cosine":
          d_norm = data.norm(dim=0, keepdim=True) + 1e-8
          c_norm = centroids.norm(dim=0, keepdim=True) + 1e-8
          data.div_(d_norm)
          centroids.div_(c_norm)
        # maxsims, labels = self.max_sim_cuda(data.transpose(-1, -2), centroids, dim=1)
        maxsims, labels = self.max_sim_cuda(
          data,
          centroids,
          dim=1,
          mode="tn"
        )
        if self.distance == "cosine":
          data.mul_(d_norm)
          centroids.mul_(c_norm)
      elif data.device.type == "cpu":
        ## doing in seperate chunks
        n_partitions = 1
        for i in range(16):
          sub_m = math.ceil(m / n_partitions)
          if self.distance == "euclidean":
            required = (sub_m*n + max(sub_m, n)) * data.element_size() + m*8 # +sub_m*d*4
          elif self.distance in ["cosine", "inner"]:
            required = (sub_m*n + sub_m+n) * data.element_size() + m*8# +sub_m*d*4
          if required < remaining:
            break
          n_partitions *= 2
        sub_m = math.ceil(m / n_partitions)
        maxsims = torch.zeros(m, device=data.device, dtype=torch.float)
        labels = torch.zeros(m, device=data.device, dtype=torch.long)
        for i in range(n_partitions):
          start = i*sub_m
          if start > m:
            break
          end = (i+1)*sub_m
          if end > m:
            end = m
          sub_data = torch.narrow(data, dim=1, start=start, length=end-start) #[d, sub_m]
          # sub_data = data[:, start:end] #[d, sub_m]
          sub_sims = self.sim(sub_data, centroids, inplace=True) #[sub_m, n]
          del sub_data
          sub_maxsims, sub_labels = sub_sims.max(dim=-1) #[sub_m]
          del sub_sims
          labels[start:end] = sub_labels
          maxsims[start:end] = sub_maxsims
          del sub_labels
      return (maxsims, labels)

  def compute_centroids_loop(self, data, labels):
    ### Naive method with loop
    d = data.shape[0]
    centroids = torch.zeros(d, self.n_clusters, device=data.device)
    unique_labels, counts = labels.unique(return_counts=True)
    for i, count in zip(unique_labels, counts):
      centroids[:, i] = data[:, labels==i].sum(dim=1) / count
    return centroids

  def compute_centroids(self, data, labels):
    """
      data: torch.Tensor, shape : [d_vector, n_data]
      labels: torch.Tensor, shape : [n_data]
      return: torch.Tensor, shape: [d_vector, n_clusters]
    """
    if data.device == torch.device("cpu"):
      centroids = self.compute_centroids_loop(data, labels)
    else:
      centroids = self.compute_centroids_cuda(data[None,], labels[None,], k=self.n_clusters)
      centroids = centroids[0]
    return centroids

  def _compute_centroids_hungry(self, data, labels):
    ### Memory hungry method
    # expanded_labels = labels[None].expand(self.n_clusters, -1) #[k, n], k=n_clusters <>
    if self.arange is None\
    or self.arange.dtype != data.dtype\
    or self.arange.device != data.device:
      self.arange = torch.arange(self.n_clusters, device=data.device) #[k] <k*8>
    
    mask = labels[None, :] == self.arange[:, None] #[k, n] <k*n*1 + k*8>
    mask_sum = mask.sum(dim=-1) #[k] <k*n*1 + k*12>
    mask = mask.float() # <k*n*5 + k*12> LARGEST MEMORY USE!!!
    centroids = mask @ data # <k*n*4 + k*12 + k*d*4>
    del mask

    centroids.div_(mask_sum[..., :, None]) # <k*d*4 + k*12>
    del mask_sum
    nan_mask = centroids!=centroids #[k, d] # <k*d*8>
    centroids[nan_mask] = 0 # remove NaNs

    return centroids

  def fit_minibatch(self, data, centroids=None):
    """
      Perform a single iteration of Minibatch KMeans algorithm
      Parameters:
        data:
          torch.Tensor
          shape : [d_vector, n_data]

        centroids: optional
          torch.Tensor
          shape : [d_vector, n_clusters]
    """
    data = data.contiguous()
    assert data.shape[1] >= self.n_clusters, "number of samples in minibatch should be greater than number of clusters"
    if centroids is not None:
      assert centroids.shape[0] == data.shape[0]
      del self.centroids
      self.register_buffer("centroids", centroids)
    elif self.centroids is None:
      new_centroids = self.initialize_centroids(data)
      del self.centroids
      self.register_buffer("centroids", new_centroids)
    else:
      assert data.shape[0] == self.centroids.shape[0]

    if self.n_points_in_clusters is None:
      n_points_in_clusters = torch.ones(self.n_clusters, device=data.device, dtype=torch.float32)
      del self.n_points_in_clusters
      self.register_buffer("n_points_in_clusters", n_points_in_clusters)
      self._iteration = 0
    
    maxsims, labels = self.get_labels(data, self.centroids)
    c_delta = self.compute_centroids(data, labels)
    weights = 1 / self.n_points_in_clusters[None]
    centroids = self.centroids * (1 - weights) + c_delta * weights
    self.centroids[:] = centroids

    unique_labels, counts = labels.unique(return_counts=True)
    self.n_points_in_clusters[unique_labels] += counts

    self._inertia = self.calculate_inertia(maxsims).item()
    self._error = self.calculate_error(centroids, self.centroids).item()
    self.print_message(f"----iteration {self._iteration}, error={self._error}, inertia={self._inertia}", 1)
    self._iteration += 1
    return labels

  def predict(self, query):
    """
      Predict closest cluster center each sample in query belongs to.
      query: torch.Tensor, shape : [d_vector, n_query]
    """
    assert self.centroids is not None, "kmeans is not trained"
    _, labels = self.get_labels(query, self.centroids)
    return labels

  def topk(self, query, k=128):
    """
      Predict the top-k closest cluster centers of each sample in query
      query: torch.Tensor, shape : [d_vector, n_query]
      k: int, should be in range [1, n_centroids]
    """
    assert self.centroids is not None, "kmeans is not trained"
    assert k <= self.n_clusters, "k is larger than number of clusters"
    if k == 1:
      # topk_v, topk_i = self.max_sim_cuda(
      #   query.transpose(-1, -2),
      #   self.centroids,
      #   dim=1
      # )
      topk_v, topk_i = self.max_sim_cuda(
        query,
        self.centroids,
        dim=1,
        mode="tn"
      )
      return (topk_v[..., None], topk_i[..., None])
    elif k <= 128:
      return self.topk_sim_cuda(
        query.transpose(-1, -2),
        self.centroids,
        dim=1,
        k=k
      )
    elif k > 128:
      sims = self.sim(query, self.centroids) #[n_query, n_clusters]
      topk_v, topk_i = sims.topk(dim=-1, k=k) #[n_query, k]
      return (topk_v, topk_i)


from abc import ABC, abstractmethod

class BaseIndex(ABC):
  def __init__(self):
    super().__init__()
  
  @abstractmethod
  def add(self):
    pass

  @abstractmethod
  def remove(self):
    pass

  @abstractmethod
  def search(self):
    pass


import torch
import cupy as cp
import numpy as np
import math
from .CustomKernel import CustomKernel
from ..util import get_absolute_path

class MBMMCuda(CustomKernel): 
  def __init__(self,
      patch_m = 4,
      patch_n = 4,
      share_mask=False,
    ):
    super(MBMMCuda, self).__init__()
    assert type(share_mask) == bool
    self.patch_m = patch_m
    self.patch_n = patch_n
    self.share_mask = share_mask    
    with open(get_absolute_path("kernels", "cuda", "mbmm.cu"),'r') as f: ###
      self.kernel = helpers + f.read()
      
    self.kernel = (self.kernel
      .replace("_PM_", str(self.patch_m))
      .replace("_PN_", str(self.patch_n))
      .replace("__MASK_BID__", "0" if share_mask else "bid")
    )
    
    self._fn_tt = cp.RawKernel(
      code=self.kernel,
      name="mbmm_tt",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )
    self._fn_nn = cp.RawKernel(
      code=self.kernel,
      name="mbmm_nn",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )
    self._fn_tn = cp.RawKernel(
      code=self.kernel,
      name="mbmm_tn",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )
    self._fn_nt = cp.RawKernel(
      code=self.kernel,
      name="mbmm_nt",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )

  def _call_nn(self, A, B, block_mask, thread_mask, element_mask):
    assert A.shape[0] == B.shape[0]
    assert A.shape[2] == B.shape[1]
    assert A.device.type == "cuda"
    assert B.device.type == "cuda"
    assert A.dtype in (torch.float, torch.half)
    assert B.dtype in (torch.float, torch.half)
    
    l, m, k = A.shape
    l, k, n = B.shape
    assert block_mask.dtype == torch.uint8 ###
    assert thread_mask.dtype == torch.uint8 ###
    assert element_mask.dtype == torch.uint8
    if self.share_mask:
      assert block_mask.shape == (math.ceil(m / 128), math.ceil(n / 128))  ###
      assert thread_mask.shape == (math.ceil(m / 8), math.ceil(n / 8)) ###
      assert element_mask.shape == (m, n)
    else:
      assert block_mask.shape == (l, math.ceil(m / 128), math.ceil(n / 128))  ###
      assert thread_mask.shape == (l, math.ceil(m / 8), math.ceil(n / 8)) ###
      assert element_mask.shape == (l, m, n)
    
    C = torch.zeros(l, m, n, device="cuda:0", dtype=A.dtype)

    threads_per_block = (256,)

    # blocks_per_grid = (l, math.ceil(n/128), math.ceil(m/128))
    n_ = math.ceil(n / (128 * self.patch_n))
    m_ = math.ceil(m / (128 * self.patch_m))
    blocks_per_grid = (self.patch_n * self.patch_m, n_ * m_, l)


    self._fn_nn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        A.data_ptr(),
        B.data_ptr(),
        C.data_ptr(),
        block_mask.data_ptr(),
        thread_mask.data_ptr(),
        element_mask.data_ptr(),
        m, n, k
      ],
      stream=self.stream
    )
    return C

  def _call_tt(self, A, B, block_mask, thread_mask, element_mask):
    raise NotImplementedError

  def _call_tn(self, A, B, block_mask, thread_mask, element_mask):
    raise NotImplementedError

  def _call_nt(self, A, B, block_mask, thread_mask, element_mask):
    raise NotImplementedError

  def __call__(
      self,
      A,
      B,
      block_mask,
      thread_mask,
      element_mask,
      mode="nn"
    ):
    """
      Performs C = f(A) @ g(B)
      A:
        torch.Tensor
        shape : [m, k] or [k, m] or [l, m, k] or [l, k, m]
        dtype : float32

      B:
        torch.Tensor
        shape : [n, k] or [k, n] or [l, n, k] or [l, k, n]
        dtype : float32

      element_mask:
        mask of elements in C that are not computed
        torch.Tensor, dtype : uint8
        if *share_mask* == True
          shape : [m, n]
        else
          shape : [l, m, n]

      block_mask:
        mask of 128x128 blocks in C that are not computed
        torch.Tensor
        dtype : uint8
        if *share_mask* == True
          shape : [ceil(m/128), ceil(n/128)]
        else
          shape : [l, ceil(m/128), ceil(n/128)]

      thread_mask:
        mask of 8x8 blocks in C that are not computed
        torch.Tensor
        dtype : uint8
        if *share_mask* == True
          shape : [ceil(m/8), ceil(n/8)]
        else
          shape : [l, ceil(m/8), ceil(n/8)]

      mode: {"nn", "tn", "nt", "tt"}, default: "nn"

      returns C:
        torch.Tensor
        shape : [m, n] or [l, m, n]
        dtype : float32

      Notes:
        f() and g() are determined by *mode*
        "nn" --> A @ B
        "tt" --> A.T @ B.T
        "nt" --> A @ B.T
        "tn" --> A.T @ B
    """
    assert len(A.shape) == len(B.shape)
    A = A.contiguous()
    B = B.contiguous()
    if len(A.shape) == 2 and len(B.shape) == 2:
      A2 = A[None]
      B2 = B[None]
      if not self.share_mask:
        block_mask = block_mask[None]
        thread_mask = thread_mask[None]
        element_mask = element_mask[None]
    elif len(A.shape) == 3 and len(B.shape) == 3:
      A2 = A
      B2 = B
    else:
      raise ValueError("shape of A and B need to be 2d or 3d")

    if mode == "nn":
      C = self._call_nn(A2, B2, block_mask, thread_mask, element_mask)
    elif mode == "tt":
      C = self._call_tt(A2, B2, block_mask, thread_mask, element_mask)
    elif mode == "tn":
      C = self._call_tn(A2, B2, block_mask, thread_mask, element_mask)
    elif mode == "nt":
      C = self._call_nt(A2, B2, block_mask, thread_mask, element_mask)

    if len(A.shape) == 2 and len(B.shape) == 2:
      C = C[0]
    return C


import torch
import cupy as cp
import numpy as np
import math
from .CustomKernel import CustomKernel
from ..util import get_absolute_path

class MinBMMCuda(CustomKernel): 
  def __init__(self, patch_m=4, patch_n=4, distance="inner"):
    super(MinBMMCuda, self).__init__()
    self.patch_m = patch_m
    self.patch_n = patch_n
    if distance == "inner":
      dist_fn = "madd"
    elif distance == "negative_inner":
      dist_fn = "nmadd"
    elif distance in ["l2", "euclidean"]:
      dist_fn = "squared_l2"
    elif distance in ["nl2", "negativel2", "negative_euclidean"]:
      dist_fn = "squared_l2"
    elif distance in ["l1", "manhattan"]:
      dist_fn = "l1"
    elif distance in ["nl1", "negativel1","negative_manhattan"]:
      dist_fn = "negative_l1"
    else:
      ValueError("Unrecognized distance type")

    self.distance = distance
    
    with open(get_absolute_path("kernels", "cuda", "bmm_helpers.cu"), "r") as f:
      helpers = f.read()
    
    with open(get_absolute_path("kernels", "cuda", "minbmm.cu"),'r') as f: ###
      self.kernel = helpers + f.read()
      
    self.kernel = (self.kernel
      .replace("_PM_", str(self.patch_m))
      .replace("_PN_", str(self.patch_n))
      .replace("__DISTANCE_FN__", dist_fn)
    )
    
    self._fn_tt = cp.RawKernel(
      code=self.kernel,
      name="min_bmm_tt",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )
    self._fn_nn = cp.RawKernel(
      code=self.kernel,
      name="min_bmm_nn",
      backend='nvcc',
      options=(
        '--maxrregcount=128',
        '--use_fast_math',
        #'-Xptxas',
        #'-dlcm=cg',
      )
    )
    # print(self._fn_nn.attributes)
    self._fn_tn = cp.RawKernel(
      code=self.kernel,
      name="min_bmm_tn",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )
    self._fn_nt = cp.RawKernel(
      code=self.kernel,
      name="min_bmm_nt",
      backend='nvcc',
      options=('--maxrregcount=128', '--use_fast_math')
    )

  def get_mode(self, A, B):
    mode = [None, None]
    if A.stride()[-1] == 1:
      mode[0] = "n"
    elif A.stride()[-2] == 1:
      mode[0] = "t"
    if B.stride()[-1] == 1:
      mode[1] = "n"
    elif B.stride()[-2] == 1:
      mode[1] = "t"
    return "".join(mode)

  def __call__(self, A, B, dim=1):
    """
      Performs C = min(f(A) @ g(B)), argmin(f(A) @ g(B))
      A: torch.Tensor, shape : [l, m, k]
      B: torch.Tensor, shape : [l, k, n]
      returns C: torch.Tensor, shape : [l, m, n]
    """
    assert len(A.shape) == len(B.shape)
    if len(A.shape) == 2 and len(B.shape) == 2:
      A = A[None]
      B = B[None]
      dim += 1
      two_dimentional = True
    elif len(A.shape) == 3 and len(B.shape) == 3:
      two_dimentional = False
    else:
      raise ValueError("A and B need to be 2d or 3d")
    assert A.shape[0] == B.shape[0]
    assert A.shape[2] == B.shape[1]
    assert A.dtype == B.dtype
    assert A.dtype in [torch.float, torch.half]
    assert A.device.type == B.device.type == "cuda"
    assert dim in [1, 2]

    mode = self.get_mode(A, B)
    
    if mode == "nn":
      kernel_fn = self._fn_nn
    elif mode == "nt":
      kernel_fn = self._fn_nt
    elif mode == "tn":
      kernel_fn = self._fn_tn
    elif mode == "tt":
      kernel_fn = self._fn_tt

    l, m, k = A.shape
    l, k, n = B.shape

    if dim == 1:
      values = torch.empty([l, n], device="cuda:0", dtype=A.dtype)
      indices = torch.empty([l, n], device="cuda:0", dtype=torch.int64)
    elif dim == 2:
      values = torch.empty([l, m], device="cuda:0", dtype=A.dtype)
      indices = torch.empty([l, m], device="cuda:0", dtype=torch.int64)
    values.fill_(float("inf"))

    threads_per_block = (256,)
    #blocks_per_grid = (math.ceil(n/128), math.ceil(m/128), l)
    
    n_ = math.ceil(n/(128*self.patch_n))
    m_ = math.ceil(m/(128*self.patch_m))
    blocks_per_grid = (self.patch_n*self.patch_m, n_ * m_, l)
    # print(blocks_per_grid, m_, n_)

    kernel_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        A.data_ptr(),
        B.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        m, n, k, dim,
      ],
      stream=self.stream
    )

    if two_dimentional:
      indices = indices[0]
      values = values[0]

    return values, indices


import torch
from .BaseContainer import BaseContainer
from .. import util

class FlatContainer(BaseContainer):
  def __init__(
      self,
      code_size,
      contiguous_size = 1,
      dtype = "float32",
      device = "cpu",
      initial_size = None,
      expand_step_size = 1024,
      expand_mode = "double",
      use_inverse_id_mapping = False,
      verbose = 0,
    ):
    super(FlatContainer, self).__init__(
      device=device,
      initial_size=initial_size,
      expand_step_size=expand_step_size,
      expand_mode=expand_mode,
      use_inverse_id_mapping=use_inverse_id_mapping
    )

    if type(dtype) == str:
      dtype = util.str2dtype(dtype)
    assert code_size > 0
    assert code_size % contiguous_size == 0

    self.code_size = code_size
    self.contiguous_size = contiguous_size
    self.dtype = dtype
    self.verbose = verbose
    self._n_items = 0

    _storage = torch.zeros(
      code_size // contiguous_size,
      self.initial_size,
      contiguous_size,
      dtype=dtype,
      device=device
    )
    self.register_buffer("_storage", _storage)

  @property
  def n_items(self):
    return self._n_items

  def get_data_by_address(self, address):
    assert util.check_dtype(address, torch.int64)
    address = address.to(self.device).clone()

    n_address = address.shape[0]
    mask = (0 <= address) & (address < self.capacity)
    address[~mask] = 0
    data = self._storage.index_select(
      dim=1,
      index=address
    )
    data.index_fill_(
      dim=1,
      index=torch.nonzero(~mask).squeeze(1),
      value=0
    )
    # data = torch.zeros(
    #   self.code_size,
    #   n_address,
    #   device=self.device,
    #   dtype=self.dtype
    # )
    # data[:, mask] = self._storage[:, address[mask]]
    data = data.transpose(1, 2).reshape(self.code_size, -1)
    return data

  def set_data_by_address(self, address, data):
    assert util.check_dtype(address, torch.int64)
    assert util.check_dtype(data, self.dtype)
    assert data.shape[0] == self.code_size
    assert data.shape[1] == address.shape[0]
    address = address.to(self.device)
    data = data.to(self.device)
    data = data.reshape(
      self.code_size // self.contiguous_size,
      self.contiguous_size
      -1,
    ).transpose(1, 2)

    mask = (0 <= address) & (address < self.capacity)
    self._storage[:, address[mask]] = data[:, mask]

  def empty(self):
    super(FlatContainer, self).empty()
    self._storage.fill_(0)
    self._n_items = 0

  def expand(self):
    super(FlatContainer, self).expand()
    n_new = self.expand_step_size
    
    # expand storage
    _storage = self._storage
    del self._storage
    new_storage = torch.zeros(
      self.code_size // self.contiguous_size,
      n_new,
      self.contiguous_size,
      device=self.device,
      dtype=self.dtype
    )
    _storage = torch.cat([
      _storage,
      new_storage
    ], dim=1)
    self.register_buffer("_storage", _storage)

  def add(self, data, ids=None, return_address=False):
    assert util.check_dtype(data, self.dtype)
    assert data.shape[0] == self.code_size
    data = data.to(self.device)
    n_data = data.shape[1]

    if ids is not None:
      assert util.check_dtype(ids, torch.int64)
      assert ids.shape[0] == n_data
      ids = ids.to(self.device)
    else:
      ids = torch.arange(
        n_data,
        device=self.device,
        dtype=torch.int64
      ) + self.max_id + 1
    
    # expand if necessary
    while (self.capacity - self.n_items) < n_data:
      self.expand()

    # add data to storage
    write_address = torch.arange(
      n_data,
      device=self.device,
      dtype=torch.int64
    ) + self.n_items
    self.set_data_by_address(data=data, address=write_address)
    self._address2id[write_address] = ids

    ### OR
    # self._storage[:, self.n_items:self.n_items+n_data] = data
    # self._address2id[self.n_items:self.n_items+n_data] = ids
    
    if self.use_inverse_id_mapping:
      self.create_inverse_id_mapping()
    self._n_items += n_data
    
    if return_address:
      return (ids, write_address)
    else:
      return ids

  def remove(self, ids=None, address=None):
    if ids is not None:
      address = self.get_address_by_id(ids)
    elif address is not None:
      address = address.to(self.device)
      # util.check_device(address, self.device)
      util.check_dtype(address, torch.int64)
    else:
      raise RuntimeError("Need either ids or address")

    mask = (0 <= address) & (address < self.capacity)
    address = address[mask].unique(sorted=True)
    n_removed = address.shape[0]
    assert n_removed <= self.n_items
    if n_removed == 0:
      return 0

    self._address2id[address] = -1
    tail_data = self._storage[:,self.n_items-n_removed:self.n_items, :]
    tail_ids = self._address2id[self.n_items-n_removed:self.n_items]
    mask = tail_ids != -1
    n = (~mask).sum().item() # number of removed items in tail
    self._storage[:, address[:n_removed-n]] = tail_data[:, mask].clone()
    self._storage[:, self.n_items-n_removed:self.n_items] = 0

    self._address2id[address[:n_removed-n]] = tail_ids[mask].clone()
    self._address2id[self.n_items-n_removed:self.n_items] = -1
    if self.use_inverse_id_mapping:
      self.create_inverse_id_mapping()
    self._n_items -= n_removed
    return n_removed



#@title BaseContainer
import torch
from abc import ABC, abstractmethod
from .. import util
from ..kernels import GetAddressByIDCuda
from ..CustomModule import CustomModule

class BaseContainer(CustomModule, ABC):
  def __init__(
      self,
      device="cpu",
      initial_size=None,
      expand_step_size=1024,
      expand_mode="double",
      use_inverse_id_mapping=False
    ):
    super(BaseContainer, self).__init__()
    if initial_size is None:
      initial_size = expand_step_size
    assert expand_mode in ["step", "double"]
    assert initial_size >= 0
    assert expand_step_size > 0

    self.device = device
    self.initial_size = initial_size
    self.expand_step_size = expand_step_size
    self.expand_mode = expand_mode
    self.use_inverse_id_mapping = use_inverse_id_mapping
    self._max_id = -1

    _address2id = torch.ones(
      initial_size,
      device=device,
      dtype=torch.long
    ) * -1
    self.register_buffer("_address2id", _address2id)
    self.register_buffer("_id2address", None)

    self._get_address_by_id_cuda = GetAddressByIDCuda(
      tpb=256
    )

  @property
  def capacity(self):
    return self._address2id.shape[0]

  @property
  def max_id(self):
    return self._max_id
    # return self._address2id.max().item() 

  def empty(self):
    self._address2id.fill_(-1)
    del self._id2address
    self.register_buffer("_id2address", None)

  def get_id_by_address(self, address):
    assert util.check_dtype(address, torch.int64)
    address = address.to(self.device)

    mask = (0 <= address) & (address < self.capacity)
    ids = torch.ones_like(address) * -1
    ids[mask] = self._address2id[address[mask]]
    return ids

  def _get_address_by_id_cpu(self, ids):
    n_ids = ids.shape[0]
    address = torch.zeros(n_ids, device=self.device, dtype=torch.long)
    for i in range(n_ids):
      id = ids[i]
      adr = torch.nonzero(self._address2id == id)
      if adr.shape[0] > 0:
        address[i] = adr[0, 0]
      else:
        address[i] = -1
    return address

  def get_address_by_id(self, ids):
    assert util.check_dtype(ids, torch.int64)
    ids = ids.to(self.device)

    if self.use_inverse_id_mapping:
      if self._id2address is None:
        self.create_inverse_id_mapping()

      # assume ids are non-negative
      mask = (0 <= ids) & (ids <= self.max_id)
      address = torch.ones_like(ids) * -1
      address[mask] = self._id2address[ids[mask]]
    else:
      if ids.device.type == "cpu":
        address = self._get_address_by_id_cpu(ids)
      elif ids.device.type == "cuda":
        address = self._get_address_by_id_cuda(self._address2id, ids)
      else:
        raise RuntimeError(f"Unrecognized device {ids.device.type}")
    return address

  def create_inverse_id_mapping(self):
    del self._id2address
    a2i_v, a2i_i = self._address2id.sort()
    a2i_mask = a2i_v >= 0
    _id2address = torch.ones(
      self.max_id + 1,
      device=self.device,
      dtype=torch.long
    ) * -1
    _id2address[a2i_v[a2i_mask]] = a2i_i[a2i_mask]
    self.register_buffer("_id2address", _id2address)

  def expand(self):
    if self.expand_mode == "double":
      self.expand_step_size *= 2
    
    _address2id = self._address2id
    del self._address2id
    new_a2i = torch.ones(
      self.expand_step_size,
      device=self.device,
      dtype=torch.long
    ) * -1
    _address2id = torch.cat([
      _address2id,
      new_a2i
    ], dim=0)
    self.register_buffer("_address2id", _address2id)

  @abstractmethod
  def add(self):
    pass

  @abstractmethod
  def remove(self):
    pass


from ..kernels import IVFPQTopkCuda
from ..kernels import IVFPQTop1Cuda

class IVFPQTopk:
  def __init__(
      self,
      n_subvectors,
      contiguous_size=4,
      sm_size=48*1024,
    ):
    self.n_subvectors = n_subvectors
    self.contiguous_size = contiguous_size
    self.sm_size = sm_size

    self._top1024_cuda = IVFPQTopkCuda(
      m=n_subvectors,
      tpb=1024,
      n_cs=contiguous_size,
      stack_capacity = 2,
      sm_size=n_subvectors * 1024,
    )

    self._top512_cuda = IVFPQTopkCuda(
      m=n_subvectors,
      tpb=512,
      n_cs=contiguous_size,
      stack_capacity = 2,
      sm_size=n_subvectors * 1024,
    )

    self._top256_cuda = IVFPQTopkCuda(
      m=n_subvectors,
      tpb=256,
      n_cs=contiguous_size,
      stack_capacity = 2,
      sm_size=n_subvectors * 1024,
    )

    if n_subvectors <= 32:
      self._top1_cuda = IVFPQTop1Cuda(
        m=n_subvectors,
        tpb=512,
        n_cs=contiguous_size,
        sm_size=n_subvectors * 1024,
      )
    else:
      self._top1_cuda = IVFPQTop1Cuda(
        m=n_subvectors,
        tpb=256,
        n_cs=contiguous_size,
        sm_size=n_subvectors * 1024,
      )

  def topk(
      self,
      data,
      precomputed,
      cell_start,
      cell_size,
      is_empty,
      n_probe_list=None,
      k=256
    ):
    assert 0 < k <= 1024
    if n_probe_list is not None:
      if k == 1:
        return self._top1_cuda.top1_smart_probing(
          data=data,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 1 < k <= 256:
        return self._top256_cuda.topk_smart_probing(
          data=data,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 256 < k <= 512:
        return self._top512_cuda.topk_smart_probing(
          data=data,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 512 < k <= 1024:
        return self._top1024_cuda.topk_smart_probing(
          data=data,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
    else:
      if k == 1:
        return self._top1_cuda.top1(
          data=data,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 1 < k <= 256:
        return self._top256_cuda.topk(
          data=data,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 256 < k <= 512:
        return self._top512_cuda.topk(
          data=data,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 512 < k <= 1024:
        return self._top1024_cuda.topk(
          data=data,
          precomputed=precomputed,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
  
  def topk_residual(
      self,
      data,
      precomputed,
      cell_start,
      cell_size,
      base_sims,
      is_empty,
      n_probe_list=None,
      k=256
    ):
    assert 0 < k <= 1024
    if n_probe_list is not None:
      if k == 1:
        return self._top1_cuda.top1_residual_smart_probing(
          data=data,
          precomputed=precomputed,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 1 < k <= 256:
        return self._top256_cuda.topk_residual_smart_probing(
          data=data,
          precomputed=precomputed,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 256 < k <= 512:
        return self._top512_cuda.topk_residual_smart_probing(
          data=data,
          precomputed=precomputed,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 512 < k <= 1024:
        return self._top1024_cuda.topk_residual_smart_probing(
          data=data,
          precomputed=precomputed,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
    else:
      if k == 1:
        return self._top1_cuda.top1_residual(
          data=data,
          precomputed=precomputed,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 1 < k <= 256:
        return self._top256_cuda.topk_residual(
          data=data,
          precomputed=precomputed,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 256 < k <= 512:
        return self._top512_cuda.topk_residual(
          data=data,
          precomputed=precomputed,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 512 < k <= 1024:
        return self._top1024_cuda.topk_residual(
          data=data,
          precomputed=precomputed,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )

  def topk_residual_precomputed(
      self,
      data,
      part1,
      part2,
      cell_start,
      cell_size,
      cells,
      base_sims,
      is_empty,
      n_probe_list=None,
      k=256
    ):
    assert 0 < k <= 1024
    if n_probe_list is not None:
      if k == 1:
        return self._top1_cuda.top1_residual_precomputed_smart_probing(
          data=data,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 1 < k <= 256:
        return self._top256_cuda.topk_residual_precomputed_smart_probing(
          data=data,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 256 < k <= 512:
        return self._top512_cuda.topk_residual_precomputed_smart_probing(
          data=data,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
      elif 512 < k <= 1024:
        return self._top1024_cuda.topk_residual_precomputed_smart_probing(
          data=data,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_probe_list=n_probe_list,
          n_candidates=k
        )
    else:
      if k == 1:
        return self._top1_cuda.top1_residual_precomputed(
          data=data,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 1 < k <= 256:
        return self._top256_cuda.topk_residual_precomputed(
          data=data,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 256 < k <= 512:
        return self._top512_cuda.topk_residual_precomputed(
          data=data,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )
      elif 512 < k <= 1024:
        return self._top1024_cuda.topk_residual_precomputed(
          data=data,
          part1=part1,
          part2=part2,
          cells=cells,
          base_sims=base_sims,
          cell_start=cell_start,
          cell_size=cell_size,
          is_empty=is_empty,
          n_candidates=k
        )



import torch

def cosine_similarity(a, b, normalize=True, inplace=False):
  """
    Compute batched cosine similarity between 'a' and 'b'
    a: torch.Tensor, shape : [l, d_vector, m]
    b: torch.Tensor, shape : [l, d_vector, n]
    normalize: bool, default : True
      if True, a and b will be normalized to norm = 1
    inplace: bool, default : False
    returns: torch.Tensor, shape : [l, m, n]
  """
  if normalize:
    a_norm = a.norm(dim=-2, keepdim=True) + 1e-8 #[l, m] <l*(m*4)>
    b_norm = b.norm(dim=-2, keepdim=True) + 1e-8 #[l, n] <l*(m*4 + n*4)>
    if inplace:
      # memory consump: m + n + (m * n)
      a.div_(a_norm)
      b.div_(b_norm)
    else:
      # memory consum: m + n + (m * n) + m*d + n*d
      a = a / a_norm #[l, d_vector, m], l*(<m*4 + n*4> + <m*d*4>)
      b = b / b_norm #[l, d_vector, n], l*(<m*4 + n*4> + <(m+n)*d*4>)
  prod = a.transpose(-2, -1) @ b #[l, m, n], <m*n*4 + m*4 + n*4> + <(m+n)*d*4>
  if inplace and normalize:
    a.mul_(a_norm)
    b.mul_(b_norm)
  return prod

def negative_squared_l2_distance(a, b, inplace=False, use_tensor_core=False):
    """
      Compute batched negative squared euclidean (l2) distance between 'a' and 'b'
      a: torch.Tensor, shape : [l, d_vector, m]
      b: torch.Tensor, shape : [l, d_vector, n]
      inplace: bool, default : False
      returns: torch.Tensor, shape : [l, m, n]
    """
    # peak mem usage: m*n*4 + max(m,n)*4 + inplace ? 0: (m+n)*d*4
    if use_tensor_core:
      y = a.transpose(-2, -1).half() @ b.half()
      y = y.float()
    else:
      y = a.transpose(-2, -1) @ b # [m, n] <m*n*4>
    y.mul_(2)
    if inplace:
      a.pow_(2)
      b.pow_(2)
    else:
      a = a ** 2 #[m, d], <m*n*4 + m*d*4>
      b = b ** 2 #[n, d], <m*n*4 + n*d*4 + m*d*4>
    a2 = a.sum(dim=-2)[..., :, None] #? [m], <m*n*4 + m*4> + <n*d*4 + m*d*4>
    y.sub_(a2)
    del a2
    b2 = b.sum(dim=-2)[..., None, :] #[n], <m*n*4 + n*4> + <n*d*4 + m*d*4>
    y.sub_(b2)
    if inplace:
      a.sqrt_()
      b.sqrt_()
    return y


from os import path
import torch

def get_absolute_path(*relative_path):
  relative_path = path.join(*relative_path)
  return path.join(path.dirname(__file__), relative_path)

def str2dtype(dtype_str):
  if dtype_str in ["double", "float64"]:
    dtype = torch.float64
  elif dtype_str in ["half", "float16"]:
    dtype = torch.float16
  elif dtype_str in ["float", "float32"]:
    dtype = torch.float32
  elif dtype_str in ["bfloat16"]:
    dtype = torch.bfloat16
  elif dtype_str in ["long", "int64"]:
    dtype = torch.int64
  elif dtype_str in ["int", "int32"]:
    dtype = torch.int32
  elif dtype_str in ["int16"]:
    dtype = torch.int16
  elif dtype_str in ["int8"]:
    dtype = torch.int8
  elif dtype_str in ["uint8"]:
    dtype = torch.uint8
  elif dtype_str in ["bool"]:
    dtype = torch.bool
  else:
    raise TypeError(f"Unrecognized dtype string: {dtype_str}")
  return dtype

def check_device(tensor, *device):
  device = [torch.device(i) if type(i) == str else i for i in device]
  return tensor.device in device

def normalize(x, dim=0):
  """
    normalize given tensor along given dimention
  """
  x_norm = x.norm(dim=dim, keepdim=True) + 1e-9
  return x / x_norm

def get_compute_capability(device_id=0):
  """
    return compute capability of GPU device
  """
  if torch.cuda.is_available():
    # device_id = torch.cuda.current_device()
    gpu_properties = torch.cuda.get_device_properties(device_id)
    result = (gpu_properties.major, gpu_properties.minor)
  else:
    result = (-1, 0)
  return result

def get_maximum_shared_memory_bytes(device_id=0):
  cc = get_compute_capability(device_id)
  if cc[0] < 7 or cc == (7, 2):
    y = 48
  elif cc == (7, 0):
    y = 96
  elif cc == (7, 5):
    y = 64
  elif cc == (8, 0):
    y = 163
  elif cc == (8, 6):
    y = 99
  else:
    y = 0
  return y * 1024

def check_dtype(tensor, *dtype):
  dtype = [str2dtype(i) if type(i) == str else i for i in dtype]
  return tensor.dtype in dtype


import torch
from ..CustomModule import CustomModule
from abc import ABC, abstractmethod

class BaseCodec(CustomModule, ABC):
  def __init__(self):
    super(BaseCodec, self).__init__()
    self.register_buffer("_is_trained", torch.tensor(False))

  def _trained(self, value):
    assert type(value) == bool
    self._is_trained.data = torch.tensor(value)

  @property
  def is_trained(self):
    return self._is_trained.item()

  @abstractmethod
  def train(self):
    pass

  @abstractmethod
  def encode(self):
    pass

  @abstractmethod
  def decode(self):
    pass
    


import torch
import cupy as cp
import math

from .CustomKernel import CustomKernel
from ..util import get_absolute_path

class IVFPQTop1Cuda(CustomKernel):
  def __init__(
      self,
      m=8,
      k=256,
      tpb=256,
      n_cs=4,
      sm_size=48*256*4,
    ):
    super(IVFPQTop1Cuda, self).__init__()
    assert tpb >= 32
    assert tpb == self.next_power_of_2(tpb), "tpb needs to be a power of 2"
    assert k == 256
    assert m * 1024 <= sm_size
    assert m % n_cs == 0
    self.m = m
    self.k = k
    self.tpb=tpb
    self.n_cs = n_cs
    self.sm_size = sm_size
    
    with open(get_absolute_path("kernels", "cuda", "ivfpq_top1.cu"), "r") as f:
      self.kernel = f.read()
    varnames = ", ".join([f"d{i}" for i in range(n_cs)])
    kernel = (self.kernel
      # .replace("_CODEBLOCK_", codeblock)
      .replace("_VARNAMES_", varnames)
      .replace("_M_", str(m))
      .replace("_K_", str(k))
      .replace("_TPB_", str(self.tpb))
      .replace("_NCS_", str(n_cs))
    )
    # print(kernel.split('\n')[60:64])
    self._top1_fn = cp.RawKernel(
      code = kernel,
      name = 'ivfpq_top1',
      options = (
        '--maxrregcount=255',
        '--use_fast_math'
      ),
      # backend='nvcc',
    )
    self._top1_fn.max_dynamic_shared_size_bytes = sm_size

    self._top1_residual_fn = cp.RawKernel(
      code = kernel,
      name = 'ivfpq_top1_residual',
      options = (
        '--maxrregcount=255',
        '--use_fast_math'
      ),
      # backend='nvcc',
    )
    self._top1_residual_fn.max_dynamic_shared_size_bytes = sm_size
    
    self._top1_residual_precomputed_fn = cp.RawKernel(
      code = kernel,
      name = 'ivfpq_top1_residual_precomputed',
      options = (
        '--maxrregcount=255',
        '--use_fast_math'
      ),
      # backend='nvcc',
    )
    self._top1_residual_precomputed_fn.max_dynamic_shared_size_bytes = sm_size

    self._top1_smart_probing_fn = cp.RawKernel(
      code = kernel,
      name = 'ivfpq_top1_smart_probing',
      options = (
        '--maxrregcount=255',
        '--use_fast_math'
      ),
      # backend='nvcc',
    )
    self._top1_smart_probing_fn.max_dynamic_shared_size_bytes = sm_size

    self._top1_residual_smart_probing_fn = cp.RawKernel(
      code = kernel,
      name = 'ivfpq_top1_residual_smart_probing',
      options = (
        '--maxrregcount=255',
        '--use_fast_math'
      ),
      # backend='nvcc',
    )
    self._top1_residual_smart_probing_fn.max_dynamic_shared_size_bytes = sm_size
    
    self._top1_residual_precomputed_smart_probing_fn = cp.RawKernel(
      code = kernel,
      name = 'ivfpq_top1_residual_precomputed_smart_probing',
      options = (
        '--maxrregcount=255',
        '--use_fast_math'
      ),
      # backend='nvcc',
    )
    self._top1_residual_precomputed_smart_probing_fn.max_dynamic_shared_size_bytes = sm_size
    

  @staticmethod
  def next_power_of_2(x):
    return 1 if x == 0 else 2**math.ceil(math.log2(x))
  
  def top1(
      self, data, precomputed,
      is_empty, cell_start, cell_size, n_candidates=1):
    """
      data: shape=[n_subvectors // n_cs, n_data, n_cs], dtype=uint8
      precomputed: shape=[n_subvectors, n_query, n_clusters], dtype=float32
      is_empty: shape=[n_data], dtype=uint8
      cell_start: shape=[n_query, n_probe], dtype=int32
      cell_size: shape=[n_query, n_probe], dtype=int32
    """
    n_data = data.shape[1]
    n_query = precomputed.shape[1]
    n_probe = cell_start.shape[1]
    assert precomputed.shape[0] == self.m
    assert precomputed.shape[2] == self.k
    assert data.shape[0] == self.m // self.n_cs
    assert data.shape[2] == self.n_cs
    assert is_empty.shape[0] == n_data
    assert cell_size.shape[1] == n_probe
    assert data.dtype == torch.uint8
    assert precomputed.dtype == torch.float32
    assert cell_start.dtype == cell_size.dtype == torch.int64
    assert is_empty.dtype == torch.uint8
    assert n_candidates == 1

    tot_size = cell_size.sum(dim=1)
    values = torch.empty(n_query, 1, device="cuda:0", dtype=torch.float32)
    values.fill_(float("-inf"))
    indices = torch.zeros(n_query, 1, device="cuda:0", dtype=torch.int64)
    threads_per_block = (self.tpb,)
    blocks_per_grid = (n_query,)

    self._top1_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.sm_size,
      args=[
        data.data_ptr(),
        precomputed.data_ptr(),
        is_empty.data_ptr(),
        cell_start.data_ptr(),
        cell_size.data_ptr(),
        tot_size.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        n_data, n_query, n_probe
        ],
      stream=self.stream
    )
    return (values, indices)

  def top1_residual(
      self, data, precomputed, base_sims,
      is_empty, cell_start, cell_size, n_candidates=1):
    """
      data: shape=[n_subvectors // n_cs, n_data, n_cs], dtype=uint8
      precomputed: shape=[n_query, n_probe, n_subvectors, n_clusters], dtype=float32
      base_sims: shape=[n_query, n_probe], dtype=float32
      is_empty: shape=[n_data], dtype=uint8
      cell_start: shape=[n_query, n_probe], dtype=int32
      cell_size: shape=[n_query, n_probe], dtype=int32
    """
    n_data = data.shape[1]
    n_query = cell_start.shape[0]
    n_probe = cell_start.shape[1]
    assert precomputed.shape == (n_query, n_probe, self.m, self.k)
    assert base_sims.shape == (n_query, n_probe)
    assert data.shape == (self.m // self.n_cs, n_data, self.n_cs)
    assert is_empty.shape == (n_data, )
    assert cell_size.shape == (n_query, n_probe)
    assert cell_start.shape == (n_query, n_probe)
    assert data.dtype == torch.uint8
    assert precomputed.dtype == torch.float32
    assert cell_start.dtype == cell_size.dtype == torch.int64
    assert is_empty.dtype == torch.uint8
    assert base_sims.dtype == torch.float32
    assert n_candidates == 1
    base_sims = base_sims.contiguous()

    tot_size = cell_size.sum(dim=1)
    values = torch.empty(n_query, 1, device="cuda:0", dtype=torch.float32)
    values.fill_(float("-inf"))
    indices = torch.zeros(n_query, 1, device="cuda:0", dtype=torch.int64)
    threads_per_block = (self.tpb,)
    blocks_per_grid = (n_query,)

    self._top1_residual_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.sm_size,
      args=[
        data.data_ptr(),
        precomputed.data_ptr(),
        base_sims.data_ptr(),
        is_empty.data_ptr(),
        cell_start.data_ptr(),
        cell_size.data_ptr(),
        tot_size.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        n_data, n_query, n_probe
        ],
      stream=self.stream
    )
    return (values, indices)

  def top1_residual_precomputed(
      self, data, part1, part2, cells, base_sims,
      is_empty, cell_start, cell_size, n_candidates=1):
    """
      data: shape=[n_subvectors // n_cs, n_data, n_cs], dtype=uint8
      part1: shape=[n_query, n_subvectors, n_pq_clusters], dtype=float32
      part2: shape=[n_cells, n_subvectors, n_pq_clusters], dtype=float32
      cells: shape=[n_query, n_probe], dtype=int64
      base_sims: shape=[n_query, n_probe], dtype=float32
      is_empty: shape=[n_data], dtype=uint8
      cell_start: shape=[n_query, n_probe], dtype=int32
      cell_size: shape=[n_query, n_probe], dtype=int32
    """
    n_data = data.shape[1]
    n_query = cell_start.shape[0]
    n_probe = cell_start.shape[1]
    assert base_sims.shape == (n_query, n_probe)
    assert cells.shape == (n_query, n_probe)
    assert data.shape == (self.m // self.n_cs, n_data, self.n_cs)
    assert is_empty.shape == (n_data, )
    assert cell_size.shape == (n_query, n_probe)
    assert cell_start.shape == (n_query, n_probe)
    assert data.dtype == torch.uint8
    assert cell_start.dtype == cell_size.dtype == torch.int64
    assert is_empty.dtype == torch.uint8
    assert base_sims.dtype == torch.float32
    assert cells.dtype == torch.int64
    assert part1.dtype == part2.dtype == torch.float32
    assert n_candidates == 1
    part1 = part1.contiguous()
    part2 = part2.contiguous()
    cells = cells.contiguous()
    base_sims = base_sims.contiguous()

    tot_size = cell_size.sum(dim=1)
    values = torch.empty(n_query, 1, device="cuda:0", dtype=torch.float32)
    values.fill_(float("-inf"))
    indices = torch.zeros(n_query, 1, device="cuda:0", dtype=torch.int64)
    threads_per_block = (self.tpb,)
    blocks_per_grid = (n_query,)

    self._top1_residual_precomputed_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.sm_size,
      args=[
        data.data_ptr(),
        part1.data_ptr(),
        part2.data_ptr(),
        cells.data_ptr(),
        base_sims.data_ptr(),
        is_empty.data_ptr(), 
        cell_start.data_ptr(),
        cell_size.data_ptr(),
        tot_size.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        n_data, n_query, n_probe
        ],
      stream=self.stream
    )
    return (values, indices)

  def top1_smart_probing(
      self, data, precomputed,
      is_empty, cell_start, cell_size, n_probe_list, n_candidates=1):
    """
      data: shape=[n_subvectors // n_cs, n_data, n_cs], dtype=uint8
      precomputed: shape=[n_subvectors, n_query, n_clusters], dtype=float32
      is_empty: shape=[n_data], dtype=uint8
      cell_start: shape=[n_query, max_n_probe], dtype=int32
      cell_size: shape=[n_query, max_n_probe], dtype=int32
      n_probe_list: shape=[n_query], dtype=int64
    """
    n_data = data.shape[1]
    n_query = precomputed.shape[1]
    n_probe = cell_start.shape[1]
    assert precomputed.shape[0] == self.m
    assert precomputed.shape[2] == self.k
    assert data.shape[0] == self.m // self.n_cs
    assert data.shape[2] == self.n_cs
    assert is_empty.shape[0] == n_data
    assert cell_size.shape[1] == n_probe
    assert data.dtype == torch.uint8
    assert precomputed.dtype == torch.float32
    assert cell_start.dtype == cell_size.dtype == torch.int64
    assert is_empty.dtype == torch.uint8
    assert n_probe_list.shape == (n_query, )
    assert n_probe_list.dtype == torch.int64
    assert n_candidates == 1

    tot_size = cell_size.sum(dim=1)
    values = torch.empty(n_query, 1, device="cuda:0", dtype=torch.float32)
    values.fill_(float("-inf"))
    indices = torch.zeros(n_query, 1, device="cuda:0", dtype=torch.int64)
    threads_per_block = (self.tpb,)
    blocks_per_grid = (n_query,)

    self._top1_smart_probing_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.sm_size,
      args=[
        data.data_ptr(),
        precomputed.data_ptr(),
        is_empty.data_ptr(),
        cell_start.data_ptr(),
        cell_size.data_ptr(),
        tot_size.data_ptr(),
        n_probe_list.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        n_data, n_query, n_probe
        ],
      stream=self.stream
    )
    return (values, indices)

  def top1_residual_smart_probing(
      self, data, precomputed, base_sims,
      is_empty, cell_start, cell_size, n_probe_list, n_candidates=1):
    """
      data: shape=[n_subvectors // n_cs, n_data, n_cs], dtype=uint8
      precomputed: shape=[n_query, max_n_probe, n_subvectors, n_clusters], dtype=float32
      base_sims: shape=[n_query, max_n_probe], dtype=float32
      is_empty: shape=[n_data], dtype=uint8
      cell_start: shape=[n_query, max_n_probe], dtype=int32
      cell_size: shape=[n_query, max_n_probe], dtype=int32
      n_probe_list: shape=[n_query], dtype=int64
    """
    n_data = data.shape[1]
    n_query = cell_start.shape[0]
    n_probe = cell_start.shape[1]
    assert precomputed.shape == (n_query, n_probe, self.m, self.k)
    assert base_sims.shape == (n_query, n_probe)
    assert data.shape == (self.m // self.n_cs, n_data, self.n_cs)
    assert is_empty.shape == (n_data, )
    assert cell_size.shape == (n_query, n_probe)
    assert cell_start.shape == (n_query, n_probe)
    assert data.dtype == torch.uint8
    assert precomputed.dtype == torch.float32
    assert cell_start.dtype == cell_size.dtype == torch.int64
    assert is_empty.dtype == torch.uint8
    assert base_sims.dtype == torch.float32
    assert n_candidates == 1
    assert n_probe_list.shape == (n_query, )
    assert n_probe_list.dtype == torch.int64
    base_sims = base_sims.contiguous()

    tot_size = cell_size.sum(dim=1)
    values = torch.empty(n_query, 1, device="cuda:0", dtype=torch.float32)
    values.fill_(float("-inf"))
    indices = torch.zeros(n_query, 1, device="cuda:0", dtype=torch.int64)
    threads_per_block = (self.tpb,)
    blocks_per_grid = (n_query,)

    self._top1_residual_smart_probing_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.sm_size,
      args=[
        data.data_ptr(),
        precomputed.data_ptr(),
        base_sims.data_ptr(),
        is_empty.data_ptr(),
        cell_start.data_ptr(),
        cell_size.data_ptr(),
        tot_size.data_ptr(),
        n_probe_list.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        n_data, n_query, n_probe
        ],
      stream=self.stream
    )
    return (values, indices)

  def top1_residual_precomputed_smart_probing(
      self, data, part1, part2, cells, base_sims,
      is_empty, cell_start, cell_size, n_probe_list, n_candidates=1):
    """
      data: shape=[n_subvectors // n_cs, n_data, n_cs], dtype=uint8
      part1: shape=[n_query, n_subvectors, n_pq_clusters], dtype=float32
      part2: shape=[n_cells, n_subvectors, n_pq_clusters], dtype=float32
      cells: shape=[n_query, max_n_probe], dtype=int64
      base_sims: shape=[n_query, max_n_probe], dtype=float32
      is_empty: shape=[n_data], dtype=uint8
      cell_start: shape=[n_query, max_n_probe], dtype=int32
      cell_size: shape=[n_query, max_n_probe], dtype=int32
      n_probe_list: shape=[n_query], dtype=int64
    """
    n_data = data.shape[1]
    n_query = cell_start.shape[0]
    n_probe = cell_start.shape[1]
    assert base_sims.shape == (n_query, n_probe)
    assert cells.shape == (n_query, n_probe)
    assert data.shape == (self.m // self.n_cs, n_data, self.n_cs)
    assert is_empty.shape == (n_data, )
    assert cell_size.shape == (n_query, n_probe)
    assert cell_start.shape == (n_query, n_probe)
    assert data.dtype == torch.uint8
    assert cell_start.dtype == cell_size.dtype == torch.int64
    assert is_empty.dtype == torch.uint8
    assert base_sims.dtype == torch.float32
    assert cells.dtype == torch.int64
    assert part1.dtype == part2.dtype == torch.float32
    assert n_candidates == 1
    assert n_probe_list.shape == (n_query, )
    assert n_probe_list.dtype == torch.int64
    part1 = part1.contiguous()
    part2 = part2.contiguous()
    cells = cells.contiguous()
    base_sims = base_sims.contiguous()

    tot_size = cell_size.sum(dim=1)
    values = torch.empty(n_query, 1, device="cuda:0", dtype=torch.float32)
    values.fill_(float("-inf"))
    indices = torch.zeros(n_query, 1, device="cuda:0", dtype=torch.int64)
    threads_per_block = (self.tpb,)
    blocks_per_grid = (n_query,)

    self._top1_residual_precomputed_smart_probing_fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.sm_size,
      args=[
        data.data_ptr(),
        part1.data_ptr(),
        part2.data_ptr(),
        cells.data_ptr(),
        base_sims.data_ptr(),
        is_empty.data_ptr(), 
        cell_start.data_ptr(),
        cell_size.data_ptr(),
        tot_size.data_ptr(),
        n_probe_list.data_ptr(),
        values.data_ptr(),
        indices.data_ptr(),
        n_data, n_query, n_probe
        ],
      stream=self.stream
    )
    return (values, indices)



#@title CellContainer
import torch
from .. import util
from ..kernels import GetDivByAddressV2Cuda
from ..kernels import GetIOACuda
from ..kernels import GetWriteAddressV2Cuda
from .BaseContainer import BaseContainer

class CellContainer(BaseContainer):
  def __init__(
      self,
      code_size,
      n_cells,
      dtype="float32",
      device="cpu",
      initial_size=None,
      expand_step_size=1024,
      expand_mode="double",
      use_inverse_id_mapping=False,
      contiguous_size=1,
      verbose=0
    ):
    if initial_size is None:
      initial_size = expand_step_size

    super(CellContainer, self).__init__(
      device = device,
      initial_size = initial_size * n_cells,
      expand_step_size = expand_step_size,
      expand_mode = expand_mode,
      use_inverse_id_mapping = use_inverse_id_mapping
    )
    assert n_cells > 0
    assert code_size > 0
    assert code_size % contiguous_size == 0
    if type(dtype) == str:
      dtype = util.str2dtype(dtype)
    self.n_cells = n_cells
    self.code_size = code_size
    self.dtype = dtype
    self.contiguous_size = contiguous_size
    self.initial_size = initial_size
    self.verbose = verbose

    _storage = torch.zeros(
      code_size // contiguous_size,
      n_cells * initial_size,
      contiguous_size,
      device = device,
      dtype = dtype
    )
    self.register_buffer("_storage", _storage)

    _cell_start = torch.arange(
      n_cells,
      device = device
    ) * initial_size
    self.register_buffer("_cell_start", _cell_start)

    _cell_size = torch.zeros(
      n_cells,
      device = device,
      dtype = torch.long
    )
    self.register_buffer("_cell_size", _cell_size)

    _cell_capacity = torch.zeros(
      n_cells,
      device = device,
      dtype = torch.long
    ) + initial_size
    self.register_buffer("_cell_capacity", _cell_capacity)

    _is_empty = torch.ones(
      n_cells * initial_size,
      device = device,
      dtype=torch.uint8
    )
    self.register_buffer("_is_empty", _is_empty)

    self._get_cell_by_address_cuda = GetDivByAddressV2Cuda(
      ta=4,
      tpb=256,
    )
    self._get_ioa_cuda = GetIOACuda(
      tpb=256,
    )
    self._get_write_address_cuda = GetWriteAddressV2Cuda(
      tpb=256,
    )

  @property
  def n_items(self):
    return self._cell_size.sum().item()

  def _get_cell_by_address_cpu(self, address, cell_start, cell_end):
    n_address = address.shape[0]
    mask1 = cell_start[None, ] <= address[:, None]
    mask2 = cell_end[None, ] > address[:, None] # [n_address, n_cq_clusters]
    mask = mask1 & mask2
    not_found = mask.sum(dim=1) == 0
    mask[not_found, 0] = True
    cells = torch.nonzero(mask)
    cells[not_found, 1] = -1
    return cells[:, 1]

  def get_cell_by_address(self, address):
    assert util.check_dtype(address, torch.int64)
    address = address.to(self.device)
    cell_start = self._cell_start
    cell_end = cell_start + self._cell_capacity
    if address.device.type == "cuda":
      return self._get_cell_by_address_cuda(address, cell_start, cell_end)
    elif address.device.type == "cpu":
      return self._get_cell_by_address_cpu(address, cell_start, cell_end)

  def _get_ioa_cpu(self, cells, unique_cells=None):
    if unique_cells is None:
      unique_cells = torch.unique(cells) #[n_unique_clusters]
    expanded_cells = cells[:, None].expand(-1, unique_cells.shape[0]) #[n_data, n_unique_clusters]
    mask = expanded_cells == unique_cells[None, :] #[n_data, n_unique_clusters]
    mcs = mask.cumsum(dim=0)
    mcs[[~mask]] = 0
    ioa = mcs.sum(dim=1) - 1
    return ioa

  def get_ioa(self, cells, unique_cells=None):
    assert util.check_dtype(cells, torch.int64)
    cells = cells.to(self.device)
    if unique_cells is not None:
      assert util.check_dtype(unique_cells, torch.int64)
      unique_cells = unique_cells.to(self.device)

    if cells.device.type == "cuda":
      return self._get_ioa_cuda(cells, unique_cells)
    elif cells.device.type == "cpu":
      return self._get_ioa_cpu(cells, unique_cells)

  def _get_write_address_cpu(self, empty_adr, cells, ioa):
    n_cells = cells.shape[0]
    cell_of_empty_adr = self.get_cell_by_address(empty_adr)
    write_address = torch.zeros_like(cells)
    for i in range(n_cells):
      cell_mask = cell_of_empty_adr == cells[i]
      write_adr = empty_adr[cell_mask]
      write_adr = write_adr[ioa[i] ]
      write_address[i] = write_adr
    return write_address

  def get_write_address(self, cells, empty_adr=None, ioa=None):
    assert util.check_dtype(cells, torch.int64)
    cells = cells.to(self.device)
    if empty_adr is not None:
      assert util.check_dtype(empty_adr, torch.int64)
      empty_adr = empty_adr.to(self.device)

    if ioa is not None:
      assert util.check_dtype(ioa, torch.int64)
      ioa = ioa.to(self.device)
    else:
      ioa = self.get_ioa(cells)

    if cells.device.type == "cuda":
      return self._get_write_address_cuda(
        self._is_empty,
        self._cell_start,
        self._cell_capacity,
        cells,
        ioa)
    elif cells.device.type == "cpu":
      assert empty_adr is not None, "empty_adr is required when device is cpu"
      return self._get_write_address_cpu(empty_adr, cells, ioa)

  def get_data_by_address(self, address):
    """
    address:
      torch.Tensor,
      shape : [n_data],
      dtype : int64
    
    returns data:
      torch.Tensor,
      shape : [code_size, n_data],
      dtype : self.dtype

    """
    assert util.check_dtype(address, torch.int64)
    address = address.to(self.device)

    n_address = address.shape[0]
    mask = (0 <= address) & (address < self.capacity)
    address[~mask] = 0
    data = self._storage.index_select(
      dim = 1,
      index = address
    )
    data.index_fill_(
      dim = 1,
      index = torch.nonzero(~mask).squeeze(1),
      value = 0,
    )
    # data = torch.zeros(
    #   self.code_size // self.contiguous_size,
    #   n_address,
    #   contiguous_size,
    #   device = self.device
    # )
    # data[:, mask] = self._storage[:, address[mask]]
    data = data.transpose(1, 2).reshape(self.code_size, -1)
    return data

  def set_data_by_address(self, data, address):
    """
    data:
      torch.Tensor,
      shape : [code_size, n_data]
      dtype : self.dtype

    address:
      torch.Tensor,
      shape : [n_data],
      dtype : int64
    """
    assert util.check_dtype(address, torch.int64)
    assert util.check_dtype(data, self.dtype)
    assert data.shape[0] == self.code_size
    assert data.shape[1] == address.shape[0]
    n_address = address.shape[0]
    address = address.to(self.device)
    data = data.to(self.device)
    data = data.reshape(
      self.code_size // self.contiguous_size,
      self.contiguous_size,
      -1
    ).transpose(1, 2)

    mask = (0 <= address) & (address < self.capacity)
    self._storage[:, address[mask]] = data[:, mask]

  def empty(self):
    super(CellContainer, self).empty()
    self._storage.fill_(0)
    self._cell_size.fill_(0)
    self._is_empty.fill_(1)
    self._n_items = 0
    self.print_message("index has been empties", 2)

  def expand(self, cells):
    n_cells = cells.shape[0]
    storage = self._storage
    address2id = self._address2id
    is_empty = self._is_empty
    del self._storage, self._address2id, self._is_empty

    total = 0
    for cell_index in cells:
      cell_start = self._cell_start[cell_index].item()
      cell_cap = self._cell_capacity[cell_index].item()
      cell_end = cell_start + cell_cap
      if self.expand_mode == "step":
        n_new = self.expand_step_size
      elif self.expand_mode == "double":
        n_new = cell_cap

      new_block = torch.zeros(
        self.code_size // self.contiguous_size,
        n_new,
        self.contiguous_size,
        device = self.device,
        dtype = self.dtype
      )
      storage = torch.cat([
        storage[:, :cell_end],
        new_block,
        storage[:, cell_end:]
      ], dim=1)

      new_a2i = torch.zeros(
        n_new,
        device = self.device,
        dtype = torch.int64
      ) - 1
      address2id = torch.cat([
        address2id[:cell_end],
        new_a2i,
        address2id[cell_end:]
      ], dim=0)

      new_is_empty = torch.ones(
        n_new,
        device = self.device,
        dtype=torch.uint8,
      )
      is_empty = torch.cat([
        is_empty[:cell_end],
        new_is_empty,
        is_empty[cell_end:]
      ], dim=0)

      self._cell_capacity[cell_index] += n_new
      arange = torch.arange(
        start = cell_index+1,
        end=self.n_cells
      )
      self._cell_start[arange] += n_new
      total += n_new
    self.register_buffer("_storage", storage)
    self.register_buffer("_address2id", address2id)
    self.register_buffer("_is_empty", is_empty)
    self.print_message(f"Total storage capacity is expanded by {total} for {n_cells} cells", 2)

  def add(self, data, cells, ids=None, return_address=False):
    assert util.check_dtype(data, self.dtype)
    assert util.check_dtype(cells, torch.long)
    assert data.shape[0] == self.code_size
    assert data.shape[1] == cells.shape[0]
    data = data.to(self.device)
    cells = cells.to(self.device)
    n_data = data.shape[1]

    if ids is not None:
      assert util.check_dtype(ids, torch.int64)
      assert ids.shape[0] == n_data
      ids = ids.to(self.device)
      
    else:
      ids = torch.arange(
        n_data,
        device=self.device,
        dtype=torch.int64,
      ) + self.max_id + 1
    
    unique_cells, unique_cell_counts = cells.unique(return_counts=True)
    ioa = self.get_ioa(cells, unique_cells)

    # expand storage if necessary
    while True:
      free_space = self._cell_capacity[cells] - self._cell_size[cells] - (ioa + 1)
      expansion_required = cells[free_space < 0].unique()
      # print("exp req", expansion_required)
      if expansion_required.shape[0] == 0:
        break
      self.expand(expansion_required)

    #get write address
    empty_adr = torch.nonzero(self._is_empty == 1)[:, 0] #[n_empty]
    write_address = self.get_write_address(
      empty_adr = empty_adr,
      cells = cells,
      ioa = ioa
    )
    self.set_data_by_address(data, write_address)

    self._address2id[write_address] = ids
    self._max_id = max(self._max_id, ids.max().item())
    self._is_empty[write_address] = 0

    # update number of stored items in each cell
    self._cell_size[unique_cells] += unique_cell_counts

    self.print_message(f"{n_data} new items added", 1)

    if return_address:
      return (ids, write_address)
    else:
      return ids

  def remove(self, ids=None, address=None):
    if ids is not None:
      address = self.get_address_by_id(ids)
    elif address is not None:
      address = address.to(self.device)
      util.check_dtype(address, torch.int64)
    else:
      raise RuntimeError("Need either ids or address")

    mask = (address >= 0) & (address < self.capacity)
    address = address[mask].unique(sorted=True)
    n_removed = address.shape[0]
    if n_removed <= self.n_items:
      self.print_message("no enough items in index to be removed", 1)
      return
    if n_removed == 0:
      return

    self._is_empty[address] = 1
    self._address2id[address] = -1

    cells = self.get_cell_by_address(address)
    unique_cells, counts = cells.unique(return_counts = True)
    self._cell_size[unique_cells] -= counts
    self.print_message(f"{n_removed} items has been removed", 2)



import torch
import numpy as np

from .IVFPQBase import IVFPQBase
from .IVFPQTopk import IVFPQTopk
from .PQ import PQ
from .MPQ import MPQ
from .kmeans import KMeans

class IVFMPQ:
  def __init__(self):
    pass
      


import torch
import torch.nn as nn
from torch.nn.utils.rnn import pad_sequence

from .kmeans import MultiKMeans
from .kmeans import KMeans
from .kernels import PQDecodeCUDA
from .PQ import PQ
from .CustomModule import CustomModule

class MPQ(CustomModule):
  def __init__(
    self,
    d_vector,
    n_subvectors=8,
    n_clusters=256,
    distance="euclidean",
    verbose=0,
    n_codebooks=64,
    ):
    super(MPQ, self).__init__()
    assert d_vector % n_subvectors == 0
    self.n_codebooks = n_codebooks
    self.d_vector = d_vector
    self.n_subvectors = n_subvectors
    self.d_subvector = d_vector // n_subvectors
    self.n_clusters = n_clusters
    self.distance = distance
    self.verbose = verbose
    self.group_size=512


    #codebook: [n_codebooks, n_subvectors, d_subvectors, n_clusters]
    self.register_buffer("codebook", None)

    self.kmeans = MultiKMeans(
      n_clusters = n_clusters,
      distance = distance,
      max_iter = 25,
      verbose = verbose,
    )

    self.codebook_selector = KMeans(
      n_clusters = n_codebooks,
      distance = distance,
      max_iter = 25,
      verbose = verbose,
    )

    self._decode_cuda = PQDecodeCUDA(tm=2, td=8)

  def train(self, x):
    """
      x: shape: [d_vector, n_data]
    """
    labels = self.codebook_selector.fit(x)

    # print("labels", labels.shape, labels.unique().shape )
    unique_labels = labels.unique()
    codebook = torch.zeros(
      self.n_codebooks,
      self.n_subvectors,
      self.d_subvector,
      self.n_clusters,
      device=x.device,
      dtype=torch.float32
    )

    for label in unique_labels:
      mask = labels == label
      sub_x = (
        x[:, mask]
        .reshape(self.n_subvectors, self.d_subvector, -1)
        .contiguous()
      )
      self.kmeans.fit(sub_x)
      codebook[label] = self.kmeans.centroids
    del self.codebook
    self.register_buffer("codebook", codebook)

  def encode(self, x):
    """
      returns code and codebook_index
      x: shape: [d_vector, n_data]
    """
    n_data = x.shape[1]

    labels = self.codebook_selector.predict(x)
    unique_labels, counts = labels.unique(return_counts=True)
    n_unique = unique_labels.shape[0]
    code = torch.zeros(self.n_subvectors, n_data, dtype=torch.uint8, device=self.codebook.device)
    for i in range(n_unique):
      label = unique_labels[i]
      mask = labels == label
      sub_x = (
        x[:, mask]
        .reshape(self.n_subvectors, self.d_subvector, -1)
        .contiguous()
      )
      sub_codebook = self.codebook[label].contiguous()
      _, sub_code = self.kmeans.get_labels(sub_x, sub_codebook)
      code[:, mask] = sub_code.byte()
    return (code, labels)

  @staticmethod
  def _decode_cpu(codebook, code):
    """
      code: torch.Tensor, shape : [n_subvectors, n_data], dtype : uint8
      return: torch.Tensor, shape : [d_vector, n_data], dtype : float32
    """
    n_subvectors, n_data = code.shape
    arange = torch.arange(n_subvectors)[:, None].expand(-1, n_data)
    res = codebook[arange, :, code.long()]
    res = res.transpose(1, 2).reshape(-1, n_data)
    return res

  def decode(self, code, codebook_index):
    """
      returns reconstruction of code
      code: [n_subvectors, n_data]
      codebook_index: shape : [n_data], dtype : uint8
    """
    n_data = code.shape[1]
    unique_labels, counts = codebook_index.unique(return_counts=True)
    recon = torch.zeros(
      self.d_vector,
      n_data,
      device=self.codebook.device,
      dtype=torch.float32,
    )
    for label in unique_labels:
      mask = codebook_index == label
      sub_code = code[:, mask].contiguous()
      sub_codebook = self.codebook[label].contiguous()
      if self.codebook.device.type == "cpu":
        sub_recon = self._decode_cpu(sub_codebook, sub_code)
      elif self.codebook.device.type == "cuda":
        sub_recon = self._decode_cuda(sub_codebook, sub_code)
      recon[:, mask] = sub_recon
    return recon

  def precompute_adc3(self, x, return_labels=False):
    d_vector, n_data = x.shape
    assert d_vector == self.d_vector
    
    labels = self.codebook_selector.predict(x)    
    unique_labels, counts = labels.unique(return_counts=True)
    n_unique = unique_labels.shape[0]

    precomputed = torch.zeros(
      self.n_subvectors,
      n_data,
      self.n_clusters,
      device=self.codebook.device
    )
    
    mask = labels[:, None] == unique_labels[None]
    xs = [ x[:, mask[:, i]].T for i in range(n_unique)]
    lens = [i.shape[0] for i in xs]

    padded_x = (
      pad_sequence(xs, batch_first=True)
      .transpose(-1, -2)
      .reshape(n_unique * self.n_subvectors, self.d_subvector, -1)
    )

    codebook = (
      self.codebook[unique_labels]
      .reshape(n_unique * self.n_subvectors, self.d_subvector, self.n_clusters)
    )

    pcd = self.kmeans.sim(padded_x, codebook, normalize=False)
    pcd = pcd.reshape(n_unique, self.n_subvectors, -1, self.n_clusters)
    for i, label in enumerate(unique_labels):
      sub_mask = mask[:, i]
      precomputed[:, sub_mask] = pcd[i, :, :lens[i] ]
    if return_labels:
      return precomputed, labels
    else:
      return precomputed

  def precompute_adc2(self, x, return_labels=False):
    d_vector, n_data = x.shape
    assert d_vector == self.d_vector
    
    labels = self.codebook_selector.predict(x)
    unique_labels, counts = labels.unique(return_counts=True)
    
    precomputed = torch.zeros(
      self.n_subvectors,
      n_data,
      self.n_clusters,
      device=self.codebook.device
    )
    
    mask = labels[:, None] == unique_labels[None]
    for i, label in enumerate(unique_labels):
      
      sub_mask = mask[:, i]
      sub_x = x[:, sub_mask]
      sub_x = sub_x.reshape(self.n_subvectors, self.d_subvector, -1)
      sub_codebook = self.codebook[label]
      sub_precomputed = self.kmeans.sim(sub_x, sub_codebook, normalize=False)
      precomputed[:, sub_mask] = sub_precomputed
    if return_labels:
      return precomputed, labels
    else:
      return precomputed
      
  def precompute_adc(self, x, return_labels=False):
    """
      x: shape : [d_vector, n_data]
    """
    d_vector, n_data = x.shape
    assert d_vector == self.d_vector

    labels = self.codebook_selector.predict(x) #[n_data]
    unique_labels, counts = labels.unique(return_counts=True)
    groups = counts // self.group_size
    unique_groups = groups.unique()
    precomputed = torch.zeros(
      self.n_subvectors,
      n_data,
      self.n_clusters,
      device=self.codebook.device
    )
    for group_index in unique_groups:
      group_unique_labels = unique_labels[groups == group_index]
      n_gul = group_unique_labels.shape[0]

      mask = labels[:, None] == group_unique_labels[None, :] #[n_data, n_gul]
      mask2 = mask.sum(dim=1).bool() #[n_data]
      sub_x = x[:, mask2]
      sub_labels = labels[mask2]

      sub_codebook = self.codebook[group_unique_labels] #[n_gul, n_subvectors, d_subvector, n_clusters]
      sub_codebook = sub_codebook.reshape(-1, self.d_subvector, self.n_clusters)# [n_gul*n_subvectors, d_subvector, n_clusters]
      padded_x = [sub_x[:, sub_labels == lab].T for lab in group_unique_labels]
      del sub_x, sub_labels

      len_x = [padded_x[i].shape[0] for i in range(n_gul)]
      padded_x = (
        pad_sequence(padded_x, batch_first=True) #[n_gul, max_n_sub_x, d_vector]
        .transpose(-1, -2) #[n_gul, d_vector, max_n_sub_x]
        .reshape(n_gul * self.n_subvectors, self.d_subvector, -1) 
      ) #[n_gul* n_subvectors, d_subvector, max_n_sub_x]

      sub_precomputed = self.kmeans.sim(padded_x, sub_codebook, normalize=False) #[n_gul*n_subvectors, max_n_sub_x, n_clusters]
      del sub_codebook, padded_x
      sub_precomputed = sub_precomputed.reshape(n_gul, self.n_subvectors, -1, self.n_clusters) #[n_gul,n_subvectors, max_n_sub_x, n_clusters]
      for i in range(n_gul):
        lab = group_unique_labels[i]
        subsub_precomputed = sub_precomputed[i][:, :len_x[i]] #[n_subvectors, n_subsub_x, n_clusters]
        sub_mask = mask[:, i]
        precomputed[:, sub_mask] = subsub_precomputed
      del sub_precomputed
    if return_labels:
      return precomputed, labels
    else:
      return precomputed


import torch
import numpy as np
from ..CustomModule import CustomModule
class PCA(CustomModule):
  def __init__(self, n_components):
    """
      Principle Component Analysis (PCA)
      n_components: int
        number of principle components
    """
    super(PCA, self).__init__()
    assert n_components > 0
    self.n_components = n_components
    self.register_buffer("_mean", None)
    self.register_buffer("_components", None)
    
  @staticmethod
  def covar(x, meaned=True, rowvar=True, inplace=False):
    """
      compute covariance matrix of 'x'
      x: torch.Tensor, shape : [m, n]
      meaned: bool, default : True
        if True, assume 'x' has zero mean
      rowvar: bool, default : True
        if True, assume 'm' represents n_features and 'n' represents n_samples
        if False, assume 'm' represents n_samples and 'n' represents n_features
      inplace: bool, default : False
        if meaned is False and inplace is True, mean of 'x' will be subtracted from 'x' inplace,
        and will be added back to 'x' at the end, this will prevent creating a new tensor of shape [m, n]
        with the cost of extra computation.
    """
    if x.dim() > 2:
      raise ValueError('x has more than 2 dimensions')
    if x.dim() < 2:
      x = x.view(1, -1)
    if not rowvar and x.shape[0] != 1:
      x = x.T
    fact = 1.0 / (x.shape[1] - 1)

    if not meaned:
      mean = x.mean(dim=1, keepdim=True)
      if inplace:
        x.sub_(mean)
      else:
        x = x - mean

    result = fact * (x @ x.T).squeeze()
    if inplace and not meaned:
      x.add_(mean)
    return result
  
  def train(self, x, inplace=False):
    """
      train PCA with 'x'
      x: torch.Tensor, shape : [d_vec, n_sample]
      inplace: bool, default : False
        if True, reduce the memory consumption with the cost of extra computation
    """
    assert x.shape[0] >= self.n_components

    mean = x.mean(dim=1, keepdim=True) #[d_vec, 1]
    if inplace:
      x.sub_(mean)
    else:
      x = x - mean
    x_cov = self.covar(x, rowvar=True, meaned=True)
    if inplace:
      x.add_(mean)
    eig_val, eig_vec = torch.symeig(x_cov, eigenvectors=True, upper=False)
    sorted_eig_val, sorted_index = eig_val.sort(descending=True)
    sorted_eig_vec = eig_vec[:, sorted_index]
    components = sorted_eig_vec[:, :self.n_components].T
    self.register_buffer("_components", components)
    self.register_buffer("_mean", mean)

  def encode(self, x):
    """
      reduce the dimentionality of 'x' from 'd_vec' to 'n_components' 
      x: torch.Tensor, shape : [d_vec, n_samples], dtype : float32
      return: torch.Tensor, shape : [n_components, n_samples], dtype : float32
    """
    assert self._components is not None
    assert x.shape[0] == self._components.shape[1]

    x = x - self._mean
    y = self._components @ x
    return y
  
  def decode(self, x):
    """
      reconstruct 'x' from 'n_components' dimentional space to 'd_vec' dimentional space 
      x: torch.Tensor, shape : [n_components, n_samples], dtype : float32
      return: torch.Tensor, shape : [d_vec, n_samples], dtype : float32
    """
    assert self._components is not None
    assert x.shape[0] == self._components.shape[0]

    y = self._components.T @ x
    y = y + self._mean
    return y